from __future__ import annotations

import functools
import itertools
import math
import uuid
from collections import Counter
from typing import Callable, Dict, Iterable, List, Optional, Sequence, Set, Tuple, TypeVar, Union, cast, overload

import torch
from typing_extensions import Concatenate, Never, ParamSpec

import rust_circuit.optional as op
from rust_circuit.py_utils import assert_never, check_cast

from ._rust import (
    Add,
    Circuit,
    Concat,
    ConstructError,
    Einsum,
    Index,
    IterativeMatcherIn,
    Matcher,
    MatcherIn,
    ModuleArgSpec,
    PushDownIndexEinsumNoopError,
    Rearrange,
    RearrangeSpec,
    Scalar,
    Symbol,
    TorchAxisIndex,
    add_elim_zeros,
    add_flatten_once,
    concat_elim_identity,
    distribute,
    einsum_elim_identity,
    einsum_flatten,
    einsum_flatten_once,
    einsum_of_permute_merge,
    einsum_permute_to_rearrange,
)
from ._rust import extract_add as extract_add_rust
from ._rust import (
    get_free_symbols,
    hash_tensor,
    index_elim_identity,
    make_broadcast,
    module_remove_unused_inputs,
    new_traversal,
    permute_of_einsum_merge,
    pull_concat_once_raw,
    pull_out_of_modules,
    push_down_index,
    push_down_index_once,
    rearrange_elim_identity,
    rearrange_fuse,
    traverse_until_depth,
)
from .py_utils import I, S, make_index_at

T = TypeVar("T")


def dedup_with_order(seq: Iterable[T]) -> List[T]:
    seen = set()
    out: List[T] = []
    for x in seq:
        if x in seen:
            continue
        else:
            seen.add(x)
            out.append(x)

    return out


def normalize_index(i: int, count: int):
    assert count >= 0, count
    if i >= 0:
        assert i < count, (i, count)
    else:
        assert i >= -count, (i, count)
    return i % count


Rewrite = Callable[[Circuit], Circuit]


class CircuitTypeError(TypeError):
    ...


class RewriteError(Exception):
    ...


class ImpossibleRewriteError(RewriteError):
    ...


class NoopRewriteError(RewriteError):
    ...


CircuitT = TypeVar("CircuitT", bound=Circuit)
CircuitT2 = TypeVar("CircuitT2", bound=Circuit)
P = ParamSpec("P")


def try_transform_wrapper(
    fn: Callable[Concatenate[CircuitT, P], CircuitT2],
    name: Optional[str] = None,
):
    """
    Wraps rewrites so that they noop rather than :class:`CircuitTypeError` or :class:`ImpossibleRewriteError`
    when called on inappropriate functions
    """

    @overload
    def wrapped_func(c: CircuitT, *args: P.args, **kwargs: P.kwargs) -> CircuitT2:
        ...

    @overload
    def wrapped_func(c: Circuit, *args: P.args, **kwargs: P.kwargs) -> Circuit:
        ...

    # This signature will also appear in the overloaded type definition, but it will never match the passed type so it
    # won't matter.
    def wrapped_func(c: Never, *args, **kwargs) -> Never:  # type: ignore
        try:
            return fn(c, *args, **kwargs)  # type: ignore
        except RewriteError:
            return c  # type: ignore
        except ConstructError:
            return c  # type: ignore
        except CircuitTypeError:
            return c  # type: ignore
        except TypeError:  # TODO: I don't currently see how to make thie exception more specific : (
            return c  # type: ignore
        except NotImplementedError:
            return c  # type: ignore

    if name is not None:
        wrapped_func.__name__ = name
    else:
        wrapped_func.__name__ = f"try_{fn.__name__}"

    return wrapped_func


def try_transform_wrapper_no_params(
    fn: Callable[[CircuitT], CircuitT2],
    name: Optional[str] = None,
):
    "like `try_transform_wrapper` but typechecks for overloaded functions"
    return try_transform_wrapper(fn, name=name)


@overload
def eliminate_zeros(circuit: Einsum) -> Scalar:
    ...


@overload
def eliminate_zeros(circuit: Rearrange) -> Scalar:
    ...


@overload
def eliminate_zeros(circuit: Index) -> Scalar:
    ...


@overload
def eliminate_zeros(circuit: Add) -> Union[Add, Scalar]:
    ...


def eliminate_zeros(circuit):
    """
    Applies the identities `x + 0 = x` and `x * 0 = 0` to the top level of the circuit.

    Also `reshape(zeros(...)) = zeros(new_shape)`, `index(zeros(...)) = zeros(new_shape)`.

    Raises error for types this transform doesn't apply to, or if it would noop
    """

    def is_zero(c: Circuit):
        return op.map(c.maybe_scalar(), lambda x: x.is_zero())

    circuit = check_cast((Einsum, Rearrange, Index, Add), circuit, exception=CircuitTypeError)
    if isinstance(circuit, (Einsum, Rearrange, Index)):
        if any(is_zero(c) for c in circuit.children):
            return Scalar(0.0, shape=circuit.shape, name=circuit.name)
        else:
            raise ImpossibleRewriteError
    elif isinstance(circuit, Add):
        zeros: Set[Circuit] = {k for k in circuit.children if is_zero(k)}
        if not zeros:
            raise ImpossibleRewriteError("No zeros to drop")
        return drop_from_add(circuit, zeros)
    else:
        assert_never(circuit)


# Needed because `eliminate_zeros` is overloaded
try_eliminate_zeros = try_transform_wrapper_no_params(eliminate_zeros)


def get_einsum_shape_d(circuit: Einsum):
    shape_d: Dict[int, int] = {}
    for c, axes in circuit.args:
        shape_d.update(dict(zip(axes, c.shape)))
    return shape_d


def rearrange_traced(
    node: Circuit,
    inp: List[List[int]],
    out: List[List[int]],
    sizes: List[Optional[int]],
    name_rearrange: Optional[str] = None,
    name: Optional[str] = None,
):
    """trace over repeated outputs, primarily for internal usage"""

    deduped_out = [list(t) for t in dedup_with_order([tuple(l) for l in out])]
    new = Rearrange(
        node,
        RearrangeSpec(inp, deduped_out, sizes),
        name=op.unwrap_or(name_rearrange, f"{node.name}_rearrange_for_trace"),
    )
    full_name = op.unwrap_or(name, f"{node.name}_rearrange_trace")
    if deduped_out == out:
        return new.rename(full_name).cast_rearrange()
    deduped_nums = tuple(range(len(deduped_out)))
    out_item_to_num = dict(zip([tuple(l) for l in deduped_out], deduped_nums))
    return Einsum(
        (new, deduped_nums),
        out_axes=tuple(out_item_to_num[tuple(x)] for x in out),
        name=full_name,
    )


def drop_mul_ones(circuit: Einsum, idxs: Optional[Iterable[int]] = None) -> Union[Einsum, Rearrange]:
    """
    Applies the identity `x*1 = x` to the top level of an Einsum circuit

    Raises error for other types
    """

    circuit = circuit.cast_einsum()

    def is_one(c: Circuit):
        return op.map(c.maybe_scalar(), lambda x: x.is_one())

    if len(circuit.args) <= 1:
        raise ImpossibleRewriteError("only should be applied to multiple argument einsum")

    to_remove = op.unwrap_or(
        op.map(idxs, lambda idxs_v: set(normalize_index(i, len(circuit.args)) for i in idxs_v)),
        {i for i, c in enumerate(circuit.children) if is_one(c)},
    )

    if len(to_remove) == 0:
        raise ImpossibleRewriteError("no Ones to remove")

    for i in to_remove:
        assert is_one(
            circuit.children[i]
        ), "maybe you passed idxs into `drop_mul_ones` which didn't correspond to ones?"

    removed_nums = set[int]().union(*(circuit.all_input_axes()[i] for i in to_remove))
    retained_nums = set[int]().union(*(axes for j, axes in enumerate(circuit.all_input_axes()) if j not in to_remove))
    no_longer_present = removed_nums - retained_nums

    new_args = [arg for i, arg in enumerate(circuit.args) if i not in to_remove]

    new_out_axes = list(filter(lambda x: x not in no_longer_present, circuit.out_axes))

    if len(no_longer_present) == 0:
        return Einsum(*new_args, out_axes=tuple(new_out_axes), name=circuit.name)

    new_out_axes = dedup_with_order(new_out_axes)

    # separate evolve to avoid renumbering
    new = Einsum(*new_args, out_axes=tuple(new_out_axes), name=f"{circuit.name}_drop_ones")

    shape_d = get_einsum_shape_d(circuit)

    # special broadcast using axes
    out = rearrange_traced(
        new,
        [[a] for a in new_out_axes],
        [[a] for a in circuit.out_axes],
        [(shape_d[a] if a in no_longer_present else None) for a in sorted(shape_d.keys())],
        name=circuit.name,
    )

    multiplier = 1.0
    for num in no_longer_present:
        if num not in circuit.out_axes:
            multiplier *= shape_d[num]

    if multiplier == 1.0:
        return out

    return Einsum.scalar_mul(
        out.rename(f"{circuit.name}_unscale"),
        scalar=multiplier,
        name=circuit.name,
    )


try_drop_mul_ones = try_transform_wrapper(drop_mul_ones)


def rearrange_of_const_to_const(circuit: Rearrange):
    """converts repeated constant to a constant"""

    circuit = circuit.cast_rearrange()
    node = circuit.node.cast_scalar()
    return Scalar(node.value, shape=circuit.shape, name=circuit.name)


try_rearrange_of_const_to_const = try_transform_wrapper(rearrange_of_const_to_const)


def flatten_adds_matching(c: Add, match: MatcherIn, allow_noop: bool = False) -> Add:
    """
    Applies the identity `x + (y + z) = x + y + z` to the top level of an Add circuit (when `(y + z)` matches the matcher function).

    Raises error for other types
    """
    c = c.cast_add()

    matcher = Matcher(match)

    out = []
    found_match = False
    for child in c.children:
        if isinstance(child, Add) and matcher(child):
            for grandchild in child.children:
                out.append(grandchild)
            found_match = True
        else:
            out.append(child)
    if not found_match and not allow_noop:
        raise ImpossibleRewriteError

    return Add(*out, name=c.name)


def flatten_adds(c: Add) -> Add:
    """
    Same as flatten_adds_matching, but uses all matchings it can
    """
    out = add_flatten_once(c.cast_add())
    if out is None:
        raise ImpossibleRewriteError
    return out


try_flatten_adds = try_transform_wrapper(flatten_adds)


def is_rearrange(c: Einsum) -> bool:
    """
    If this einsum has one argument and no repeated axes, it's a rearrange.
    """
    return len(c.args) == 1 and len(c.args[0][1]) == len(set(c.args[0][1])) and len(c.out_axes) == len(set(c.out_axes))


def einsum_to_rearrange(c: Einsum) -> Rearrange:
    """
    Replaces Einsum with equivalent Rearrange.
    Raises:
        ImpossibleRewriteError: If input Einsum cannot be converted to rearrange.
    """
    out = einsum_permute_to_rearrange(c.cast_einsum())

    if out is None:
        raise ImpossibleRewriteError
    return out.cast_rearrange()


try_einsum_to_rearrange = try_transform_wrapper(einsum_to_rearrange)

T = TypeVar("T")


def flatten(l: List[List[T]]) -> List[T]:
    return [x for sl in l for x in sl]


def explicit_squeeze(c: Rearrange, suffix: str = "_no_squeeze") -> Einsum:
    spec = c.spec

    # Squeeed are empty lists
    nb_squeezed_dims = len([l for l in spec.input_ints if not l])
    if nb_squeezed_dims == 0:
        raise ImpossibleRewriteError("No squeezed dimensions")

    new_spec = RearrangeSpec(spec.input_ints, spec.output_ints + [[] for _ in range(nb_squeezed_dims)], spec.int_sizes)

    all_axes = tuple(range(len(new_spec.output_ints)))

    return Einsum(
        (Rearrange(c.node, new_spec, name=f"{c.name}{suffix}"), all_axes),
        out_axes=all_axes[: c.rank],
        name=c.name,
    )


try_explicit_squeeze = try_transform_wrapper(explicit_squeeze)


def remove_noop_rearrange(c: Union[Einsum, Rearrange], keep_name: bool = False) -> Circuit:
    """
    Attempts to strip rearranges that noop at the top level of the circuit for trivial Einsums and Rearranges

    Errors for other types, ImpossibleRewriteErrors where there isn't a noop rearrange

    Note: for einsum with no args, returns 1, as in remove_empty_einsum
    Note: for einsum with with times 1, remove the times 1
    """
    c = check_cast((Einsum, Rearrange), c, exception=CircuitTypeError)

    if isinstance(c, Einsum):
        out = einsum_elim_identity(c)
        if out is not None:
            return out
        else:
            raise ImpossibleRewriteError(f"Einsum {c} is not identity")

    out = rearrange_elim_identity(c.cast_rearrange())
    if out is not None:
        if keep_name:
            return out.rename(c.name)
        else:
            return out
    else:
        raise ImpossibleRewriteError(f"Rearrange {c} is not a noop")


try_remove_noop_rearrange = try_transform_wrapper(remove_noop_rearrange)


# remove_add_times_one -> remove_add_single_summand
# maybe keep remove_add_times_one for backcompatibility?
# maybe make it deal with add of einsum.scalar_mul ?
# maybe use remove_add_few_input?
def remove_add_single_summand(c: Add, keep_name: bool = False) -> Circuit:
    """
    Strips add with only one summand at the top level of the circuit.

    Raises error for other circuit types, and :class:`ImpossibleRewriteError` for :class:`Add` s
    """
    c = c.cast_add()

    if c.num_children == 1:
        (out,) = c.children
        if keep_name:
            return out.rename(c.name)
        else:
            return out
    else:
        raise ImpossibleRewriteError("Input has more than 1 element")


try_remove_add_single_summand = try_transform_wrapper(remove_add_single_summand)


def remove_single_concat(c: Concat, keep_name: bool = False) -> Circuit:
    """
    Strips concat with only one argument the top level of the circuit.

    Raises error for other circuit types, and ImpossibleRewriteError for :class:`Add` s
    """
    c = c.cast_concat()

    out = concat_elim_identity(c)

    if out is not None:
        if keep_name:
            return out.rename(c.name)
        else:
            return out
    else:
        raise ImpossibleRewriteError("Input has more than 1 element")


try_remove_single_concat = try_transform_wrapper(remove_single_concat)

# TODO: add remove_inclusive_specified_slice back?
def remove_trivial_index(c: Index) -> Circuit:
    """
    Strips index which has only slice(None) at top level of circuit

    Raises error for other circuit types or an ImpossibleRewriteError when types are correct
    """
    c = c.cast_index()

    out = index_elim_identity(c)
    if out is not None:
        return out
    else:
        raise ImpossibleRewriteError


try_remove_trivial_index = try_transform_wrapper(remove_trivial_index)

# maybe remove? (already covered by remove_noop_rearrange)
def remove_empty_einsum(c: Einsum) -> Scalar:
    """
    Replaces empty einsum with One (or noop for other einsums)

    Raises error for other circuit types, ImpossibleRewriteError is not an empty Einsum
    """
    c = c.cast_einsum()
    if len(c.args) == 0:
        return Scalar(1.0, name=c.name)
    else:
        raise ImpossibleRewriteError


try_remove_empty_einsum = try_transform_wrapper(remove_empty_einsum)

# maybe revert to nodes: Set[Circuit]?
def drop_from_add(add: Add, drop_match: MatcherIn, noop_error: bool = False) -> Union[Add, Rearrange, Scalar]:
    """
    Not a rewrite! Remove specified summands.

    See drop_zero_add below
    """
    add = add.cast_add()
    drop_matcher = Matcher(drop_match)
    new_nodes = [node for node in add.children if not drop_matcher(node)]

    if noop_error:
        if add.num_children == 0:
            raise NoopRewriteError("`add` is empty")
        if len(new_nodes) == add.num_children:
            raise NoopRewriteError("No nodes removed")

    if len(new_nodes) == 0:
        return Scalar(0.0, add.shape, name=add.name)
    out: Circuit = Add(*new_nodes, name=add.name)

    if out.shape != add.shape:
        # TODO: fix name duplicate?
        braocasted = make_broadcast(out, add.shape)
        if braocasted is None:
            raise ImpossibleRewriteError
        out = braocasted.rename(out.name)
    return check_cast((Add, Rearrange, Scalar), out, exception=CircuitTypeError)


try_drop_from_add = try_transform_wrapper(drop_from_add)

# warning! the meaning changed! here it's node zero, not weight zero. Maybe change the name?
def drop_zero_add(c: Add) -> Add:
    """
    Applies the identity `x + 0 = x` at the top level of the circuit.

    Raises error for other types, NoopRewriteError if nothing to drop
    """
    c = c.cast_add()

    out = add_elim_zeros(c)
    if out is None:
        raise NoopRewriteError
    return out


try_drop_zero_add = try_transform_wrapper(drop_zero_add)

# maybe use remove_add_few_input?
def zero_empty_add(c: Add) -> Scalar:
    """
    Replaces add-with-no-summands with zero.

    Errors if wrong type, ImpossibleRewriteError if not applicable.
    """

    c = c.cast_add()

    if c.num_children > 0:
        raise ImpossibleRewriteError(f"{c.num_children} must be zero")
    return Scalar(0.0, shape=c.shape, name=c.name)


try_zero_empty_add = try_transform_wrapper(zero_empty_add)


def factor_add_of_mul_to_mul_of_add(c: Add, is_lhs: bool = True) -> Einsum:
    """
    Applies the identity `a*b + a*c = a(b + c)` (or `b*a + c*a = (b + c)*a`) at the top level of the circuit.

    Errors when type is wrong. Raises ImpossibleRewriteError where not applicable
    """
    c = c.cast_add()

    items = [k.cast_einsum() for k in c.children]

    dim_vals = {tuple((c.shape, axes) for c, axes in x.args) for x in items}
    out_axes_vals = {x.out_axes for x in items}

    if len(dim_vals) != 1:
        raise ImpossibleRewriteError(f"Summands have {len(dim_vals)} != 2 different axis specifications")

    out_axes = list(out_axes_vals)[0]
    dim_items = list(dim_vals)[0]

    if len(dim_items) != 2:
        raise ImpossibleRewriteError(f"Each summand contains the product of {len(dim_items)} != 2 things")

    args_vals = [{x.children[i] for x in items} for i in range(len(dim_items))]

    count_is_1 = [len(a) == 1 for a in args_vals]
    count_is_1_total = sum(count_is_1)

    if count_is_1_total < len(args_vals) - 1 or len(args_vals) <= 1:
        raise ImpossibleRewriteError  # TODO this is redundant

    if not count_is_1[int(not is_lhs)]:
        raise ImpossibleRewriteError("There isn't a common factor that can be factored out")

    if is_lhs:
        assert len(list(args_vals[0])) == 1
        lhs: Circuit = list(args_vals[0])[0]
        rhs: Circuit = Add(*[k.children[1] for k in items], name=f"{c.name}_new_add_from_factor")
    else:
        lhs = Add(*[k.children[0] for k in items], name=f"{c.name}_new_add_from_factor")
        assert len(list(args_vals[1])) == 1
        rhs = list(args_vals[1])[0]

    (_, lhs_axes), (_, rhs_axes) = list(dim_vals)[0]

    return Einsum(
        (lhs, lhs_axes),
        (rhs, rhs_axes),
        out_axes=out_axes,
        name=c.name,
    )


try_factor_add_of_mul_to_mul_of_add = try_transform_wrapper(factor_add_of_mul_to_mul_of_add)

# maybe add back more precise typing at the end
# TODO: check names are not broken
def pull_through_add_concat(c: Add) -> Circuit:
    """
    Replaces sum of concats with concat of sums.

    All concats must have children of the same shapes and be along the same axis.

    Errors for wrong types, ImpossibleRewriteError for things that are Adds but fail
    """

    out = pull_concat_once_raw(c.cast_add())
    if out is None:
        raise ImpossibleRewriteError
    return out


try_pull_through_add_concat = try_transform_wrapper(pull_through_add_concat)


def index_key(index: List[TorchAxisIndex]) -> tuple:
    """Convert index attr into a hashable tuple"""
    out = []

    for i in index:
        if isinstance(i, torch.Tensor):
            out.append(hash_tensor(i))
        elif isinstance(i, slice):
            out.append((i.start, i.stop, i.step))
        elif isinstance(i, int):
            out.append(i)
        else:
            assert_never(i)
    return tuple(out)


def pull_through_add_index(c: Add) -> Index:
    """
    Replaces sum of indexes with index of sums.

    All indexes must have the same output shape and the "same" (by hash) index attr.

    Errors for other types, ImpossibleRewriteError for things that are Adds but fail
    """

    c = c.cast_add()
    items: List[Index] = []
    for k in c.children:
        if not isinstance(k, Index):
            raise ImpossibleRewriteError(f"Child {k} is not an Index")
        items.append(k)

    shapes_index = {(k.node.shape, index_key(k.idx)) for k in items}
    if len(shapes_index) != 1:
        raise ImpossibleRewriteError

    return Index(
        Add(*[k.node for k in items], name=f"{c.name}_new_add_for_index"),
        items[0].idx,
        name=c.name,
    )


try_pull_through_add_index = try_transform_wrapper(pull_through_add_index)


def pull_through_add_rearrange(c: Add) -> Union[Add, Rearrange]:
    """
    Replaces sum of rearranges with rearrange of sum.

    All rearranges must have the same output shape and the same spec.

    Errors for other types, ImpossibleRewriteError for things that are Adds but fail
    """

    items: List[Rearrange] = []
    for k in c.children:
        if not isinstance(k, Rearrange):
            raise ImpossibleRewriteError(f"Child {k} is not a Rearrange")
        items.append(k)

    shapes_spec = set((k.node.shape, k.spec) for k in items)
    if len(shapes_spec) != 1:
        raise ImpossibleRewriteError(f"Children Rearranges have different specs, {shapes_spec}")

    ((_, spec),) = shapes_spec

    return Rearrange(
        Add(*[k.node for k in items], name=f"{c.name}_new_add_for_rearrange"),
        spec,
        name=c.name,
    )


try_pull_through_add_rearrange = try_transform_wrapper(pull_through_add_rearrange)


def apply_permute_axes_ints(
    inp: Tuple[int, ...], out: Tuple[int, ...], axes: Tuple[T, ...], is_on_output: bool
) -> Tuple[T, ...]:
    assert len(out) == len(axes)
    if is_on_output:
        # Swap the variables so we apply the permutation to output
        inp, out = out, inp
    return tuple(axes[out.index(inp)] for inp in inp)


def apply_permute_axes(spec: RearrangeSpec, axes: Tuple[T, ...], is_on_output: bool) -> Tuple[T, ...]:
    assert spec.is_permute()
    flat_inp = tuple(i for (i,) in spec.input_ints)
    flat_out = tuple(i for (i,) in spec.output_ints)
    return apply_permute_axes_ints(flat_inp, flat_out, axes, is_on_output=is_on_output)


def is_permute(c: Circuit):
    return rearrange_no_splits_or_joins(c, just_permute=True)


def has_no_splits_or_joins(spec: RearrangeSpec):
    return all([len(x) == 1 for x in spec.input_ints]) and all([len(x) == 1 for x in spec.output_ints])


def rearrange_no_splits_or_joins(c: Circuit, just_permute: bool = False):
    if not isinstance(c, Rearrange):
        return False

    spec = c.spec.canonicalize()
    if just_permute:
        return spec.is_permute()
    else:
        return has_no_splits_or_joins(spec)


def unpack_n(*args: Iterable[T], n: int) -> Tuple[Tuple[T, ...], ...]:
    out = tuple(zip(*args))
    if len(args) == 0:
        return ((),) * n
    assert len(out) == n
    return out


def einsum_end_num(c: Einsum) -> int:
    return max(itertools.chain(*c.all_input_axes(), c.out_axes), default=-1) + 1


def fuse_einsum_rearrange(
    c: Einsum, just_permute: bool = False, index: Optional[int] = None, force_despite_diaged_repeat: bool = False
) -> Union[Einsum, Rearrange]:
    """
    Replaces einsum of non splitting/joining rearrange with just einsum.

    Errors if wrong type, ImpossibleRewriteError otherwise

    By default, doesn't fuse if there is a output diag of a repeated dim (can't really be simplified, but we can put in different form)
    """

    def rearrange_info(rearrange: Rearrange):
        spec = rearrange.spec.canonicalize()

        flat_inp = tuple(i for (i,) in spec.input_ints)
        flat_out = tuple(i for (i,) in spec.output_ints)

        return spec, flat_inp, flat_out

    def rearrange_repeat_output_dims(rearrange: Rearrange):
        _, flat_inp, flat_out = rearrange_info(rearrange)

        return [i for i, o in enumerate(flat_out) if o not in flat_inp]

    c = c.cast_einsum()
    first_rearrange = next(
        iter(
            [
                i
                for i, (x, axes) in enumerate(c.args)
                if rearrange_no_splits_or_joins(x, just_permute=just_permute)
                and isinstance(x, Rearrange)
                and (
                    (not any(c.out_axes.count(axes[i]) > 1 for i in rearrange_repeat_output_dims(x)))
                    or force_despite_diaged_repeat
                )
            ]
        ),
        None,
    )
    re_idx = op.unwrap_or(index, first_rearrange)
    if re_idx is None:
        raise ImpossibleRewriteError

    re_idx = normalize_index(re_idx, len(c.args))
    assert re_idx is not None  # mypy strikes again

    (rearrange_v, rearrange_axes), other_args = get_info(c, re_idx)
    rearrange = rearrange_v.cast_rearrange()

    if not rearrange_no_splits_or_joins(rearrange):
        raise ImpossibleRewriteError

    spec, flat_inp, flat_out = rearrange_info(rearrange)
    assert len(rearrange_axes) == len(flat_out)

    assert len(set(flat_inp)) == len(flat_inp)
    assert len(set(flat_out)) == len(flat_out)

    extra_num = einsum_end_num(c)

    for j, inp in enumerate(flat_inp):
        if inp not in flat_out:
            assert spec.int_sizes[j] == 1
            flat_out += (inp,)
            rearrange_axes += (extra_num,)
            extra_num += 1

    (non_repeat_axes, non_repeat_out), (repeat_axes, repeat_out) = [
        unpack_n(*filter(lambda x: (x[1] in flat_inp) == is_in_inp, zip(rearrange_axes, flat_out)), n=2)
        for is_in_inp in [True, False]
    ]

    assert set(non_repeat_out) == set(flat_inp)

    new_axes = apply_permute_axes_ints(flat_inp, non_repeat_out, non_repeat_axes, is_on_output=False)

    out_other_args = other_args
    if len(repeat_out) != 0:
        out_other_args += (
            (
                Scalar(1.0, tuple(spec.int_sizes[i] for i in repeat_out), name=f"{c.name}_ones_from_repeat"),
                repeat_axes,
            ),
        )

    return reconstruct_by(c, re_idx, rearrange.node, new_axes, out_other_args)


try_fuse_einsum_rearrange = try_transform_wrapper(fuse_einsum_rearrange)

# maybe add back index?
def fuse_einsum_permute(c: Einsum) -> Einsum:
    """
    Replaces einsum of permute with just einsum.

    Errors if wrong type, ImpossibleRewriteError otherwise
    """
    out = einsum_of_permute_merge(c)
    if out is None:
        raise ImpossibleRewriteError
    return out


try_fuse_einsum_permute = try_transform_wrapper(fuse_einsum_permute)


def fuse_permute_einsum(c: Rearrange) -> Einsum:
    """
    Replaces permute of einsum with just einsum.
    (we could generalize to repeating also via ones, but we'll avoid this)

    Errors if wrong type, ImpossibleRewriteError otherwise
    """
    out = permute_of_einsum_merge(c)
    if out is None:
        raise ImpossibleRewriteError
    return out


try_fuse_permute_einsum = try_transform_wrapper(fuse_permute_einsum)


def fuse_rearrange(c: Rearrange) -> Rearrange:
    """
    Fuses two rearranges into a single rearrange.

    Errors if both of the rearranges contain splits or joins (fuse is not always possible in this case).
    """
    out = rearrange_fuse(c.cast_rearrange())
    if out is None:
        raise ImpossibleRewriteError(
            "Both rearranges contain a split/join"
        )  # TODO: check that rearrange fuse is still as general
    return out


try_fuse_rearrange = try_transform_wrapper(fuse_rearrange)


def fuse_single_einsum(c: Einsum) -> Einsum:
    """
    Replaces single einsum of einsum with just 1 einsum.

    Errors if wrong type, ImpossibleRewriteError otherwise
    """
    c = c.cast_einsum()
    if len(c.args) == 1 and isinstance(c.children[0], Einsum):
        out = einsum_flatten_once(c)
        if out is None:
            raise ImpossibleRewriteError
        return out
    else:
        raise ImpossibleRewriteError


try_fuse_single_einsum = try_transform_wrapper(fuse_single_einsum)


def fuse_einsum_single(circ: Einsum) -> Einsum:
    """
    Replaces einsum of single einsum with just 1 einsum.

    Errors if type is wrong and ImpossibleRewriteError if not applicable.
    """
    circ = circ.cast_einsum()
    to_not_fuse = [c for c in circ.children if not (isinstance(c, Einsum) and len(c.args) == 1)]
    if len(to_not_fuse) == len(circ.children):
        raise ImpossibleRewriteError("There are no Einsums with 1 child under this one")

    new_c = einsum_flatten(circ, new_traversal(end_depth=3, term_early_at=set(to_not_fuse)))
    return new_c


try_fuse_einsum_single = try_transform_wrapper(fuse_einsum_single)


def get_info(einsum: Einsum, idx: int):
    return einsum.args[idx], tuple(einsum.args[:idx] + einsum.args[idx + 1 :])


def reconstruct_by(
    einsum: Einsum,
    idx: int,
    x: Circuit,
    axes: Tuple[int, ...],
    other_args: Tuple[Tuple[Circuit, Tuple[int, ...]], ...],
    new_name: Optional[str] = None,
    new_out_axes: Optional[Tuple[int, ...]] = None,
):
    new_args = other_args[:idx] + ((x, axes),) + other_args[idx:]
    out = einsum.evolve(args=new_args)
    if new_out_axes is not None:
        out = out.evolve(out_axes=new_out_axes)
    if new_name is not None:
        out = cast(Einsum, out.rename(new_name))
    return out


def split_einsum_concat(
    c: Einsum,
    element_index: int,
    index_groups: List[List[int]],
    group_names: Optional[Sequence[str]] = None,
    axis_name: Optional[str] = None,
    use_axis_name: bool = True,
    push_down_through_concat: bool = True,
    apply_try_remove_single_concat: bool = True,
) -> Union[Add, Concat]:
    """
    Replaces an einsum where one input is a concat with a concat (or add) of einsums (each with smaller concats).

    Every group in index_groups must be a contiguous increasing list of indices. The groups must be ordered in
    increasing order by their smallest element. e.g. the following are valid index groups:
    ::

        [[2]]
        [[1, 2]]
        [[0], [2]]
        [[0, 1], [2]]

    If any indices are not in a group, they are automatically grouped into contiguous stretches. e.g. in the first
    example above, ``[0, 1]`` would also be a group; in the second, ``[0]``; and in the third, ``[1]``.

    Args:
        c: The einsum to rewrite.
        is_lhs: If True, splits up the left argument to the einsum (must be a concat). Otherwise splits up the right.
        index_groups: List of lists of indexes to group into concats. These refer to the order circuits appear in the
            original concat. Must be nonempty and each element must be nonempty.

    Returns:
        Concat (or sum) of einsums, one per index group. Each einsum has the original rhs (lhs) argument if is_lhs
        (not is_lhs), and its lhs (rhs) is a concat of the children of the original lhs (rhs) specified by the indices
        in the group (or just the child if the group is of size 1). Concat order is preserved.

        The output is a concat if the original concat dim was in the output of the einsum; and a sum otherwise.
    """
    c = c.cast_einsum()

    assert len(index_groups) > 0, "you must have at least one index group"
    assert all(len(g) > 0 for g in index_groups), "groups must be non-empty"

    node_v, _ = c.args[element_index]
    node = node_v.cast_concat()

    index_groups = list(sorted((list(sorted(g)) for g in index_groups), key=lambda x: x[0]))

    new = []
    running = 0
    for group in index_groups:
        assert len(group) > 0, group
        assert len(group) == len(set(group)), group
        if group[0] != running:
            assert group[0] > running, (group[0], running)
            new.append(list(range(running, group[0])))
        new.append(group)
        if len(group) != group[-1] - group[0] + 1:
            raise NotImplementedError("only supports contiguous groups atm")
        running = len(group) + group[0]

    assert running <= len(node.children), (running, node)
    if running < len(node.children):
        new.append(list(range(running, len(node.children))))

    index_groups = new
    axis = node.axis

    new_slice: List[slice] = []

    for g in index_groups:
        assert len(g) > 0, g

        before = sum(node.children[j].shape[axis] for j in range(0, g[0]))
        end = before + sum(node.children[j].shape[axis] for j in range(g[0], g[-1] + 1))

        new_slice.append(S[before:end])

    def run_on_index(new_idx: Index, _, name: str):
        if not push_down_through_concat:
            return new_idx

        try:
            pushed_down = push_down_index_once(new_idx, suffix="_" + name).cast_einsum()
        except PushDownIndexEinsumNoopError:
            # Detects identity einsum (even fancy ones not detected by elim_identity_einsum, like aa->a when shape is (1,1))
            raise ImpossibleRewriteError("Can't split_einsum_concat on identity einsum")

        (this_node_v, axes), other = get_info(pushed_down, element_index)
        if isinstance(this_node_v, Index):
            assert this_node_v.node == node_v
            new_node_v = push_down_index_once(this_node_v, suffix="").map_children(
                lambda x: remove_trivial_index(x.cast_index())
            )
        else:
            assert this_node_v == node_v
            new_node_v = this_node_v
        assert isinstance(new_node_v, Concat)
        if apply_try_remove_single_concat:
            new_node_v = try_remove_single_concat(new_node_v)

        return reconstruct_by(pushed_down, element_index, new_node_v, axes, other)

    return explicit_reduce(
        c,
        axis=axis,
        element_index=element_index,
        partitioning_idxs=new_slice,
        partitioning_idx_names=group_names,
        axis_name=axis_name,
        use_axis_name=use_axis_name,
        run_on_index=run_on_index,
    )


try_split_einsum_concat = try_transform_wrapper(split_einsum_concat)


def distribute_old(c: Einsum, element_index: int, suffix: Optional[str] = None, iters: int = 1) -> Add:
    """
    Recursively applies the identity `a*(b + c) = a*b + a*c`.

    Args:
        c: The einsum to distribute.
        element_index: The value of ``i`` in ``a_1 * a_2 * ... * (b_i + c_i) * ...``, if the Einsum has the latter
            expression.
        suffix: For naming the new terms. If None, new nodes won't be named.
        iters: Number of times to recurse into children. Must be at least 1.
    """

    c = c.cast_einsum()

    if iters <= 0:
        raise ImpossibleRewriteError("iters must be >= 0")
    if (
        element_index >= len(c.args)
        or c.args[element_index][0].maybe_add() is None
        or len(c.args[element_index][0].cast_add().children) == 0
    ):
        # Make sure at least one level of distribute if possible
        raise ImpossibleRewriteError(f"child of {c} at {element_index} must be an none empty Add")

    out = distribute(
        c,
        element_index,
        traverse_until_depth(iters),
        suffix,
        allow_partial_distribute=True,
        do_broadcasts=True,
    ).cast_add()

    return out


try_distribute_old = try_transform_wrapper(distribute_old)


def permute_to_einsum(c: Rearrange) -> Einsum:
    spec = c.cast_rearrange().spec
    if not spec.is_permute():
        raise ImpossibleRewriteError
    flat_inp = tuple(itertools.chain(*spec.input_ints))
    flat_out = tuple(itertools.chain(*spec.output_ints))

    return Einsum((c.node, flat_inp), out_axes=flat_out, name=c.name)


try_permute_to_einsum = try_transform_wrapper(permute_to_einsum)


def push_down_permute_via_einsum(c: Rearrange, suffix: Optional[str] = "perm", iters=1) -> Add:
    """similar to distribute for einsum"""
    c = c.cast_rearrange()
    return distribute_old(permute_to_einsum(c), element_index=0, suffix=suffix, iters=iters)


try_push_down_permute_via_einsum = try_transform_wrapper(push_down_permute_via_einsum)

# TODO: instead of equiv partition, we should support nest_adds!

# https://stackoverflow.com/questions/38924421/is-there-a-standard-way-to-partition-an-interable-into-equivalence-classes-given
def iterable_gen_equivalence_partition(
    iterable: Iterable[T], relation: Callable[[T, T], bool]
) -> Tuple[List[List[T]], Dict[T, List[T]]]:
    """Partitions a set of objects into equivalence classes

    Args:
        iterable: collection of objects to be partitioned
        relation: equivalence relation. I.e. relation(o1,o2) evaluates to True
            if and only if o1 and o2 are equivalent

    Returns: classes, partitions
        classes: A sequence of sets. Each one is an equivalence class
        partitions: A dictionary mapping objects to equivalence classes
    """
    classes: List[List[T]] = []
    partitions: Dict[T, List[T]] = {}
    for o in iterable:  # for each object
        # find the class it is in
        found = False
        for c in classes:
            if relation(c[0], o):  # is it equivalent to this class?
                c.append(o)
                partitions[o] = c
                found = True
                break
        if not found:  # it is in a new class
            classes.append([o])
            partitions[o] = classes[-1]
    return classes, partitions


def equivalence_partition_list(
    c: Add,
    get_group_name: Callable[[Circuit], str],
    is_eq: Callable[[Circuit, Circuit], bool],
) -> List[Add]:
    """
    Rearranges sum to a list of sums of groups of items (grouped by is_eq).

    Returned list can be summed to recover an equivalent sum to the original.
    """
    classes, _ = iterable_gen_equivalence_partition(c.children, lambda a, b: is_eq(a, b))
    return [Add(*c, name=get_group_name(c[0])) for c in classes]


def equivalence_partition(
    c: Add,
    get_group_name: Callable[[Circuit], str],
    is_eq: Callable[[Circuit, Circuit], bool],
    noop_error: bool = False,
) -> Add:
    """Rearranges sum to first sum groups of items (grouped by is_eq), then sum the groups."""
    if noop_error and len(c.children) == 0:
        raise NoopRewriteError("Add node `c` has no elements")
    return Add(*equivalence_partition_list(c, get_group_name, is_eq), name=c.name)


def make_single_explicit_reduce(i: int):
    return (I[:i] + (i,) + I[i + 1 :]), (f"before_{i}", f"{i}", f"after_{i}")


def make_single_explicit_reduce_idxed(i: int, size: int, device: Optional[Union[torch.device, str]] = None):
    return ((torch.cat([torch.arange(i), torch.arange(i + 1, size)]).to(device=device),) + (i,)), (f"not_{i}", f"{i}")


@functools.cache
def make_single_explicit_reduce_idxed_cached(i: int, size: int, device: Optional[Union[torch.device, str]] = None):
    """use at your peril: this function will fill up the RAM with non-garbage-collectable objects"""
    return make_single_explicit_reduce_idxed(i, size, device)


def explicit_reduce(
    c: Einsum,
    axis: int,
    element_index: int,
    partitioning_idxs: Sequence[TorchAxisIndex],
    partitioning_idx_names: Optional[Sequence[str]] = None,
    axis_name: Optional[str] = None,
    use_axis_name: bool = True,
    run_on_index: Callable[[Index, TorchAxisIndex, str], Circuit] = lambda x, *_: x,
    check: bool = True,
    use_dot: bool = False,
) -> Union[Add, Concat]:
    """explicitly concat with separate nodes (and then reduce) instead of implicitly via einsum

    Can be used for things like having a different circuit for each attention head.
    """
    c = c.cast_einsum()

    if len(c.args) == 0:
        raise ImpossibleRewriteError("c is an empty Einsum")

    _, node_axes = c.args[element_index]
    if len(node_axes) == 0:
        raise ImpossibleRewriteError("The asked-for child of Einsum is a scalar")

    num = node_axes[axis]

    if c.out_axes.count(num) > 1:
        raise NotImplementedError("trace case not implemented")

    is_concat = num in c.out_axes

    if use_axis_name:
        axis_name_prefix = "_" + op.unwrap_or(axis_name, "at")
    else:
        axis_name_prefix = ""

    if is_concat:
        new_out_axes = c.out_axes
        new_c = c
    else:
        new_out_axes = (num,) + c.out_axes
        # TODO improve naming maybe
        new_c = Einsum(*c.args, out_axes=new_out_axes, name=f"{c.name}{'.' if use_dot else '_'}keep{axis_name_prefix}")

    axis_loc = new_out_axes.index(num)

    vals = split_reduce_concat_impl(
        new_c,
        axis_loc=axis_loc,
        is_concat=is_concat,
        partitioning_idxs=partitioning_idxs,
        partitioning_idx_names=partitioning_idx_names,
        axis_name=axis_name,
        use_axis_name=use_axis_name,
        run_on_index=run_on_index,
        check=check,
        use_dot=use_dot,
    )

    if is_concat:
        return Concat(*vals, axis=axis_loc, name=c.name)
    else:
        assert len(partitioning_idxs) > 0
        nums = tuple(range(new_c.rank))
        return Add(
            *[
                v if isinstance(idx, int)
                # nums[1:] is right because new_out_axes = (num,) + c.out_axes
                else Einsum((v, nums), out_axes=nums[1:], name=f"{v.name}_sum")
                for v, idx in zip(vals, partitioning_idxs)
            ],
            name=c.name,
        )


def split_reduce_concat_impl(
    c: Circuit,
    axis_loc: int,
    is_concat: bool,
    partitioning_idxs: Sequence[TorchAxisIndex],
    partitioning_idx_names: Optional[Sequence[str]] = None,
    axis_name: Optional[str] = None,
    use_axis_name: bool = True,
    run_on_index: Callable[[Index, TorchAxisIndex, str], Circuit] = lambda x, *_: x,
    check: bool = True,
    use_dot: bool = False,
):
    assert axis_loc < c.rank

    if use_axis_name:
        axis_name_suffix = op.unwrap_or(axis_name, "at") + "_"
    else:
        axis_name_suffix = ""

    if partitioning_idx_names is None:
        partitioning_idx_names = [f"idx_{i}" for i in range(len(partitioning_idxs))]
    assert len(partitioning_idx_names) == len(partitioning_idxs)

    partitioning_idxs = [normalize_index(i, c.shape[axis_loc]) if isinstance(i, int) else i for i in partitioning_idxs]
    if is_concat:
        partitioning_idxs = [slice(i, i + 1) if isinstance(i, int) else i for i in partitioning_idxs]
    for idx in partitioning_idxs:
        if isinstance(idx, torch.Tensor):
            assert idx.ndim == 1

    if check:
        device = None
        for idx in partitioning_idxs:
            if isinstance(idx, torch.Tensor):
                device = idx.device

        to_set = torch.zeros((c.shape[axis_loc],), dtype=torch.bool, device=device)
        for idx in partitioning_idxs:
            assert (~to_set[idx]).any() or to_set[idx].nelement() == 0
            to_set[idx] = True
            if is_concat:
                # torch.sort on CUDA doesn't support bool
                to_set_long = to_set.long()
                assert (torch.sort(to_set_long, descending=True)[0] == to_set_long).all()
        assert to_set.all(), (to_set, to_set.shape)

    return [
        run_on_index(
            Index(
                c,
                make_index_at(p_idx, axis_loc),
                name=f"{c.name}{'.' if use_dot else '_'}{axis_name_suffix}{p_idx_name}",
            ),
            p_idx,
            f"{axis_name_suffix}{p_idx_name}",
        )
        for p_idx, p_idx_name in zip(partitioning_idxs, partitioning_idx_names)
    ]


def split_to_concat(
    c: Circuit,
    axis: int,
    partitioning_idxs: Sequence[TorchAxisIndex],
    partitioning_idx_names: Optional[Sequence[str]] = None,
    axis_name: Optional[str] = None,
    use_axis_name: bool = True,
    run_on_index: Callable[[Index, TorchAxisIndex, str], Circuit] = lambda x, *_: x,
    check: bool = True,
    use_dot: bool = False,
) -> Concat:
    """
    Replaces the circuit `c` with a concatenation of slices of `c`.
    Slices along `axis`, where the i-th slice will include the indicies `partition_idxs[i]`.

    Naming options:
    By default each Index will be named `{c.name}_at_idx_{i}`
    `use_dot` replaces the first underscore with a period
    `axis_name` replaces "at_". `use_axis_name=False` removes this part entirely.
    `partitioning_idx_names` replaces `idx_{i}`

    Other options:
    `run_on_index` is an optional transformation to run on every created index. For instance, a call to `push_down_index` could be provided.
    `check` ensures it is an algebric rewrite.

    See also: rc.split_to_concat which has less options but is in rust (and thus probably faster?).
    """
    # this exists for fancy naming options
    vals = split_reduce_concat_impl(
        c,
        axis_loc=axis,
        is_concat=True,
        partitioning_idxs=partitioning_idxs,
        partitioning_idx_names=partitioning_idx_names,
        axis_name=axis_name,
        use_axis_name=use_axis_name,
        run_on_index=run_on_index,
        check=check,
        use_dot=use_dot,
    )
    return Concat(*vals, axis=axis, name=c.name)


# TODO: port to rust completly (I assume some code is already doing sth like that?)
def extract_output_diags(c: Einsum):
    """separate einsum for output diags"""

    c = c.cast_einsum()

    if len(c.out_axes) == len(set(c.out_axes)):
        raise ImpossibleRewriteError("no output diags to extract")

    deduped_axes = tuple(dedup_with_order(c.out_axes))
    evolved = Einsum(*c.args, out_axes=deduped_axes, name=f"{c.name}_no_diag")

    return Einsum((evolved, deduped_axes), out_axes=c.out_axes, name=c.name)


try_extract_output_diags = try_transform_wrapper(extract_output_diags)


def extract_input_diags(c: Einsum, item_index: int):
    """separate einsum for input diags, nice for distributing and similar"""

    c = c.cast_einsum()

    (node, node_axes), other = get_info(c, item_index)

    deduped_axes = tuple(dedup_with_order(node_axes))

    if len(deduped_axes) == len(node_axes):
        raise ImpossibleRewriteError()

    new_node = Einsum((node, node_axes), out_axes=deduped_axes, name=f"{node.name}_traced")

    return reconstruct_by(c, item_index, new_node, deduped_axes, other)


try_extract_input_diags = try_transform_wrapper(extract_input_diags)


def einsum_flatten_bans_noop(c: Einsum, traversal: IterativeMatcherIn = new_traversal(), allow_noop: bool = False):
    """
    equivalent to einsum_flatten, but bans noop (raises exception)
    """
    out = einsum_flatten(c, traversal)
    if out == c and not allow_noop:
        raise NoopRewriteError
    return out


try_einsum_flatten = try_transform_wrapper(einsum_flatten_bans_noop)


def residual_rewrite(c: Circuit, estim: Circuit, running_name: str, flip_naming: bool = False):
    """
    Rewrites c as `(estim + (c - estim))`. Then returns the tuple:
      - estim + (c - estim),    # named `{c.name}`
      - estim,
      - c - estim               # named `{c.name}_{running_name}_residual`

    The inner copy of c is renamed to `{c.name}_orig`.

    If flip_naming is true then the entire circuit is named `{c.name}_{running_name}_out` and the inner copy of c is named `{c.name}`.
    """
    residual = c.rename(c.name if flip_naming else f"{c.name}_orig").sub(estim, f"{c.name}_{running_name}_residual")

    return (estim.add(residual, name=f"{c.name}_{running_name}_out" if flip_naming else c.name), estim, residual)


# TODO: maybe implement extract add in terms of nest_adds (which hasn't yet been implemented)


def extract_add(c: Add, sub: Add) -> Add:
    """Extract sub from c: a+b+c+d, c+d -> a+b+(c+d)"""
    out = extract_add_rust(c.cast_add(), sub.cast_add())
    if out is None:
        raise ImpossibleRewriteError
    return out


def extract_add_by_match(c: Add, match: Matcher, sub_name: Optional[str] = None):
    """Extract matching node from c: a+b+c+d -> b+c+(a+d) (where a & d match)"""
    c = c.cast_add()
    matcher = Matcher(match)
    sub = Add(*[node for node in c.children if matcher(node)], name=op.unwrap_or(sub_name, c.name + "_sub"))

    return extract_add(c, sub)


def get_dups(*elems: T) -> Set[T]:
    counts = Counter(elems)
    return set([x for x in elems if counts[x] > 1])


def check_permutation(perm: List[int], count: int, allow_rest: bool = False):
    perm_set = set(perm)
    assert len(perm) == len(perm_set), (perm, get_dups(perm))
    assert perm_set.issubset(range(count)), (perm, count, perm_set - set(range(count)))
    rest = set(range(count)) - perm_set
    if not allow_rest:
        assert len(rest) == 0, rest
    return rest


def nested_einsum_axes_permute(c: Einsum, element_index: int, permutation: Sequence[int], keep_name: bool = True):
    """TODO: maybe reimplement in terms of rearrange muls"""
    c = c.cast_einsum()

    (node_v, node_axes), other = get_info(c, element_index)
    node = node_v.cast_einsum()

    perm = list(permutation)
    check_permutation(perm, len(node_axes), allow_rest=False)

    def get_permuted(vals: Tuple[int, ...]):
        new_axes = torch.zeros((len(node_axes),), dtype=torch.long)
        new_axes[torch.tensor(perm, dtype=torch.long)] = torch.tensor(vals, dtype=torch.long)
        return tuple(int(i.item()) for i in new_axes)

    new_out_axes = get_permuted(node.out_axes)
    new_node_axes = get_permuted(node_axes)

    new_node = Einsum(*node.args, out_axes=new_out_axes, name=node.op_name if keep_name else f"{node.name}_perm")
    result = reconstruct_by(c, element_index, new_node, new_node_axes, other)
    if result == c:
        raise NoopRewriteError("nested_einsum_axes_permute noop")
    return result


def nested_einsum_permute_dups_to_eq(c: Einsum):
    """TODO: maybe reimplement in terms of rearrange muls"""
    # untested, annoying to test...
    c = c.cast_einsum()

    def norm_out(x: Einsum):
        return x.evolve(out_axes=tuple(sorted(x.out_axes)))

    to_class: Dict[Tuple[int, Einsum], List[Tuple[int, Einsum]]] = iterable_gen_equivalence_partition(
        [(i, x) for i, x in enumerate(c.children) if isinstance(x, Einsum)],
        relation=lambda x, y: norm_out(x[1]) == norm_out(y[1]),
    )[1]

    for (idx, x), ((_, first), *_) in to_class.items():
        if x != first:
            c = nested_einsum_axes_permute(
                c, idx, permutation=[x.out_axes.index(i) for i in first.out_axes], keep_name=True
            )

    return c


try_nested_einsum_permute_dups_to_eq = try_transform_wrapper(nested_einsum_permute_dups_to_eq)


def multiply_axes_by_identity(
    c: Circuit,
    axes: Sequence[int],
    ident_prefix_name: Optional[str] = None,
    keep_sub_name: bool = False,
    batch_axes: Sequence[int] = [],
):
    """note that this will be simplified away!"""
    axes = [normalize_index(a, c.rank) for a in axes]
    shape = tuple(c.shape[a] for a in batch_axes) + tuple(c.shape[a] for a in axes)
    assert len(set(batch_axes).intersection(axes)) == 0
    ident_prefix_name_v = op.unwrap_or(ident_prefix_name, c.name + "_ident")
    batch_nums = tuple(range(len(batch_axes)))
    nums = tuple(range(len(batch_nums), len(batch_nums) + len(axes)))
    ident = Einsum(
        (Scalar(1.0, shape, f"{ident_prefix_name_v}_one"), batch_nums + nums),
        out_axes=batch_nums + nums + nums,
        name=ident_prefix_name_v,
    )

    all_nums = tuple(range(c.rank))
    out_axes = tuple(range(c.rank, c.rank + len(axes)))

    all_nums_out_replaced = list(all_nums)
    for i, a in enumerate(axes):
        all_nums_out_replaced[a] = out_axes[i]

    return Einsum(
        (c.rename(c.name + ("" if keep_sub_name else "_no_ident")), all_nums),
        (ident, tuple(batch_axes) + tuple(axes) + out_axes),
        out_axes=tuple(all_nums_out_replaced),
        name=c.name + ("_ident" if keep_sub_name else ""),
    )


# TODO: test me!!!
def make_diag_mask_pair(n: int, ndim: int, name_n: str):
    diag = Einsum(
        (Scalar(1.0, (n,), name=f"{name_n}_size_one"), (0,)), out_axes=(0,) * ndim, name=f"{name_n}_diag_{ndim}"
    )
    non_diag = Scalar(1.0, (n,) * ndim, name=f"{name_n}_{ndim}_size_one").sub(diag, name=f"{name_n}_non_diag_{ndim}")

    return Add(diag, non_diag, name=f"{name_n}_one_mask_pair")


# TODO: test me!!!
def insert_diag_mul_mask(
    c: Einsum, item_idx: int, n: int, dims: Sequence[int], name_n: str, use_distribute: bool = True
):
    return insert_diag_mul_mask_many(c, [item_idx], n, [dims], name_n, use_distribute=use_distribute)


# TODO: test me!!!
def insert_diag_mul_mask_many(
    c: Einsum, item_idxs: Sequence[int], n: int, dims: Sequence[Sequence[int]], name_n: str, use_distribute: bool = True
):
    c = c.cast_einsum()
    all_nums = [c.all_input_axes()[i] for i in item_idxs]
    item_nums = tuple(itertools.chain(*([nums[i] for i in dims_per] for dims_per, nums in zip(dims, all_nums))))
    mask_pair = make_diag_mask_pair(n, len(item_nums), name_n=name_n)

    out = c.evolve(args=tuple(c.args) + ((mask_pair, item_nums),))

    if use_distribute:
        out_add = distribute_old(out, element_index=len(c.args))
        assert len(out_add.children) == 2
        return out_add.map_children(lambda x: x.rename(c.name + ("_non_diag" if "non_diag" in x.name else "_diag")))

    return out


# maybe use split_to_concat
def split_to_concat_for_batch(circ: Circuit, batch_size: int, axis: int = 0):
    assert circ.rank >= 1

    return Concat(
        *[
            Index(circ, make_index_at(I[i * batch_size : min((i + 1) * batch_size, circ.shape[axis])], axis))
            for i in range(math.ceil(circ.shape[axis] / batch_size))
        ],
        axis=axis,
    )


# useful for pushing an index symbol "past" the symbol into module arguments
def push_down_indexes_through_modules(
    circ, matcher: MatcherIn = Matcher(True), push_down: Callable[[Index], Circuit] = push_down_index
):
    syms = set(get_free_symbols(circ))
    no_parent_fvs = Matcher(lambda c: syms.isdisjoint(set(get_free_symbols(c))))

    def rename_idx(y) -> Circuit:
        if y.is_index():
            y = y.map_child(0, rename_idx)
            return y.rename(y.get_autoname())
        return y

    def update_bind(y):
        y = rename_idx(y)
        if y.is_index():
            return push_down(y.cast_index())
        return y

    m = Matcher(Index) & (lambda y: y.children[0].is_symbol()) & no_parent_fvs

    def construct(m, old):
        m2 = module_remove_unused_inputs(
            m, use_elim_no_input_module=False, add_suffix_on_remove_unused=False
        ).cast_module()
        assert m2.num_children <= module_remove_unused_inputs(old).num_children, (
            "push_down_indexes_through_modules would result in duplicating an input (index of it + it, or multiple different indexes of it)",
            m2,
        )
        return m2

    return pull_out_of_modules(
        circ,
        matcher,
        update_new_binds=update_bind,
        matcher_new_binds=m,
        module_construct_callback=construct,
        circuit_to_arg_spec=lambda c: ModuleArgSpec(
            Symbol(c.shape, uuid.UUID(bytes=c.hash[:16]), c.get_unique(Symbol).name), True, True, True
        ),
    )



import contextlib
from typing import ClassVar


class NotebookDoneTestingException(Exception):
    """
    A demo/notebook can raise this exception when __name__ != "__main__" to be tested only up to some point.
    """

    pass


class NotebookInTesting(contextlib.ContextDecorator):
    currently_in_notebook_test: ClassVar[bool] = False

    def __enter__(self):
        if self.__class__.currently_in_notebook_test:
            raise RuntimeError("NotebookInTesting context manager not nestable")
        self.__class__.currently_in_notebook_test = True
        return self

    def __exit__(self, exc_type: Exception, exc_val, exc_tb) -> bool:
        # Cannot nest this context manager correctly, because on exiting we always
        # set this variable to False.
        # To nest it properly, we'd have to remember what the value was on __enter__: too much work.
        self.__class__.currently_in_notebook_test = False
        if exc_type == NotebookDoneTestingException:
            return True  # Ignore exception
        return False

    @classmethod
    def exit_if_in_testing(cls):
        if cls.currently_in_notebook_test:
            raise NotebookDoneTestingException



import os
from typing import *

import attrs
import msgpack
import numpy
import torch

callables_table = {}


def _msgpack_special_encoder(obj):
    if isinstance(obj, numpy.ndarray):
        # Unfortunately there is no Int64Array type in Javascript.
        # So we convert int64 arrays to int32 arrays, and encode them over the wire that way.
        # However, we maintain the original type to allow for (somewhat) reversible decoding.
        # This is pretty absurdly confusing, so I might change this later.
        if obj.dtype == numpy.int64:
            obj = obj.astype(numpy.int32)
        if obj.dtype == numpy.uint64:
            obj = obj.astype(numpy.uint32)
        if obj.dtype == numpy.float32 or obj.dtype == numpy.float64:
            obj = obj.astype(numpy.float16)  # always send float16 to save bandwidth
        return {
            "$$type$$": "ndarray",
            "dtype": obj.dtype.name,
            "shape": obj.shape,
            "data": obj.tobytes(),
            "v": 0,  # I might change the over-the-wire encoding later, so I want a version tag.
        }
    if isinstance(obj, torch.Tensor):
        array_encoding = _msgpack_special_encoder(obj.detach().cpu().numpy())
        return {"$$type$$": "torch", "array": array_encoding, "device": str(obj.device)}
    if hasattr(obj, "to_dict"):
        return obj.to_dict()
    if attrs.has(obj.__class__):
        return attrs.asdict(obj)
    if callable(obj):
        function_token = os.urandom(12).hex()
        callables_table[function_token] = obj
        return {"$$type$$": "local-callback", "id": function_token}
    raise TypeError("cannot serialize %r object" % type(obj))


def _decode_walk_tree(x):
    if isinstance(x, dict):
        if "$$type$$" in x:
            if x["$$type$$"] == "ndarray":
                array = numpy.frombuffer(
                    x["data"],
                    dtype={
                        "int8": numpy.int8,
                        "uint8": numpy.uint8,
                        "int16": numpy.int16,
                        "uint16": numpy.uint16,
                        "int32": numpy.int32,
                        "uint32": numpy.uint32,
                        "int64": numpy.int32,  # Intentional mismatch! See comment above.
                        "uint64": numpy.uint32,  # Intentional mismatch! See comment above.
                        "float32": numpy.float32,
                        "float64": numpy.float64,
                    }[x["dtype"]],
                ).reshape(x["shape"])
                if x["dtype"] == "int64":
                    array = array.astype(numpy.int64)
                if x["dtype"] == "uint64":
                    array = array.astype(numpy.uint64)
                return array
            elif x["$$type$$"] == "torch":
                array = _decode_walk_tree(x["array"])
                return torch.tensor(array)
                # return torch.tensor(array, device=x["device"])
            else:
                raise ValueError("Bad serialized $$type$$: %r" % x["$$type$$"])
        return {k: _decode_walk_tree(v) for k, v in x.items()}
    if isinstance(x, list):
        return [_decode_walk_tree(v) for v in x]
    return x


def msgpack_encode(obj: Any) -> bytes:
    return msgpack.packb(obj, default=_msgpack_special_encoder)


def msgpack_decode(raw_data: bytes) -> Any:
    return _decode_walk_tree(msgpack.unpackb(raw_data))


def get_callback_result(msg):
    if msg["id"] in callables_table:
        function = callables_table[msg["id"]]
        data = msg["data"]
        value = function(*data)
        response = dict(kind="callbackResult", token=msg["token"], buffer=value)
    else:
        response = dict(kind="callbackResult", token=msg["token"], buffer="unknown callback token")
    return response



from time import perf_counter
from typing import Generic, Iterable, List, Mapping, NoReturn, Optional, Tuple, TypeVar, Union

import torch

from ._rust import TorchAxisIndex


def assert_never(x: NoReturn) -> NoReturn:
    raise AssertionError(f"Invalid value: {x!r}")


def check_cast(cls, c, exception=TypeError):
    if isinstance(c, cls):
        return c
    else:
        raise exception(f"Requested type {repr(cls)} does not match the type of the input, {type(c)}.")


class timed:
    def __init__(self, name, min_to_print=None, extra_print=None):
        self.name = name
        self.min_to_print = min_to_print
        self.extra_print = extra_print

    def __enter__(self):
        self.start = perf_counter()
        return self

    def __exit__(self, type, value, traceback):
        self.time = perf_counter() - self.start
        self.readout = f"{self.name} took {self.time:.4f}"
        if self.min_to_print is None or self.min_to_print < self.time:
            print(self.readout + (self.extra_print if self.extra_print is not None else ""))


TorchIndex = Union[Tuple[TorchAxisIndex, ...], TorchAxisIndex]


class Indexer:
    """Helper for defining slices more easily which always returns tuples
    (instead of sometimes returning just TorchAxisIndex)."""

    def __getitem__(self, idx: TorchIndex) -> Tuple[TorchAxisIndex, ...]:
        if isinstance(idx, tuple):
            return idx
        return (idx,)


class Slicer:
    """Helper for defining slices more easily which always returns slices"""

    def __getitem__(self, idx: slice) -> slice:
        assert isinstance(idx, slice)
        return idx


I = Indexer()
S = Slicer()

KT = TypeVar("KT")
VT = TypeVar("VT")


# Dict is an invariant type by default. We declare FrozenDict to be covariant using Generic.
class FrozenDict(dict, Generic[KT, VT]):
    # TODO: maybe this should actually use frozendict...
    _cached_hash: Optional[int] = None
    _cached_tuple: Optional[Tuple] = None

    def _to_tuple(self):
        if self._cached_tuple is None:
            self._cached_tuple = tuple(sorted(self.items(), key=lambda x: hash(x)))
        return self._cached_tuple

    def __hash__(self):
        if self._cached_hash is None:
            self._cached_hash = hash(("tao's FrozenDict", self._to_tuple()))
        return self._cached_hash

    def __eq__(self, other):
        # Possible hash collisions when sorting the tuple in _to_tuple make the
        # implementation of FrozenDict.__eq__ complicated"
        # TODO use frozendict
        return isinstance(other, FrozenDict) and dict(self) == dict(other)

    def clear_cache(self):
        self._cached_hash = self._cached_tuple = None

    def __repr__(self):
        return f"{self.__class__.__name__}({super().__repr__()})"


def get_slice_list(n: int) -> List[TorchAxisIndex]:
    return [slice(None)] * n


def make_index_at(index: TorchIndex, at: int):
    return make_index_at_many({at: index})


def to_axis_index(idx: TorchIndex) -> TorchAxisIndex:
    if isinstance(idx, tuple):
        assert len(idx) == 1
        return idx[0]
    else:
        return idx


def make_index_at_many(at_idx: Mapping[int, TorchIndex]) -> Tuple[TorchAxisIndex, ...]:
    out = get_slice_list(max(at_idx.keys(), default=-1) + 1)
    for at, idx in at_idx.items():
        out[at] = to_axis_index(idx)

    return tuple(out)



import torch

from interp.tools.indexer import I
from rust_circuit import (
    Add,
    Array,
    Concat,
    Einsum,
    Index,
    OptimizationSettings,
    Parser,
    Rearrange,
    RearrangeSpec,
    batch_einsum,
    batch_to_concat,
    propagate_named_axes,
)

P = Parser()


def test_named_axes():
    circs = [
        Einsum.from_einsum_string(
            "abc,acd,ade->ba",
            Array(torch.randn(128, 32, 32), "hi").rename_axes({0: "batch"}),
            Array.randn(128, 32, 32),
            Array.randn(128, 32, 32),
        ),
        Index(
            Einsum.from_einsum_string(
                "abc,acd,ade->ba",
                Array(torch.randn(128, 32, 32), "hi").rename_axes({0: "batch"}),
                Array.randn(128, 32, 32),
                Array.randn(128, 32, 32),
            ),
            I[0, :],
        ),
        Rearrange(
            Index(
                Einsum.from_einsum_string(
                    "abc,acd,ade->ba",
                    Array(torch.randn(128, 32, 32), "hi").rename_axes({0: "batch"}),
                    Array.randn(128, 32, 32),
                    Array.randn(128, 32, 32),
                ),
                I[0, :],
            ),
            RearrangeSpec([[0]], [[1], [0], [2]], [None, 2, 3]),
        ),
        Add(
            Rearrange(
                Index(
                    Einsum.from_einsum_string(
                        "abc,acd,ade->ba",
                        Array(torch.randn(128, 32, 32), "hi").rename_axes({0: "batch"}),
                        Array.randn(128, 32, 32),
                        Array.randn(128, 32, 32),
                    ),
                    I[0, :],
                ),
                RearrangeSpec([[0]], [[1], [0], [2]], [None, 2, 3]),
            ),
            Array(torch.randn(128, 3), "hi2").rename_axes({0: "batch2", 1: "rando"}),
        ),
        Concat(
            Rearrange(
                Index(
                    Einsum.from_einsum_string(
                        "abc,acd,ade->ba",
                        Array(torch.randn(128, 32, 32), "hi").rename_axes({0: "batch"}),
                        Array.randn(128, 32, 32),
                        Array.randn(128, 32, 32),
                    ),
                    I[0, :],
                ),
                RearrangeSpec([[0]], [[1], [0], [2]], [None, 2, 3]),
            ),
            Array(torch.randn(2, 128, 3), "hi2").rename_axes({1: "batch2", 2: "rando"}),
            axis=1,
        ),
    ]
    for circ in circs:
        circ.print()
        propagate_named_axes(circ, {}, True).print()


def test_batching_top():
    circs = [
        Einsum.from_einsum_string(
            "ab,bc->ac",
            Add(
                Array.randn(2, 4),
                Array.randn(2, 4),
            ),
            Array.randn(4, 8),
        ),
        P(
            """
0 Einsum ab,bc->ac
  2 Add
    3 [2, 3] Scalar 1.0
    4 [2, 3] Scalar 1.1
  5 Einsum ab,->ab
    9 [3, 5] Scalar 1.2
    6 Einsum ab->
      3
  """
        ),
        P(
            """
0 Einsum ab,bc->ac
  1 Add
    2 [2, 4] Scalar 1.0
    3 [2, 4] Scalar 1.1
  4 [4, 8] Scalar 1.2
  """
        ),
        P(
            """
'outer_b.bilinear_mlp' Module
  'bilinear_mlp' Einsum h,h,oh->o
    'bilinear_mlp.pre0' Index [0,:]
      'bilinear_mlp.fold_pre' Rearrange (a:2 b) -> a:2 b
        'bilinear_mlp.pre' Einsum i,hi->h
          'bilinear_mlp.input' [?] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52
          'bilinear_mlp.w.proj_in' [?,?] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6
    'bilinear_mlp.pre1' Index [1,:]
      'bilinear_mlp.fold_pre'
    'bilinear_mlp.w.proj_out' [?,?] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf
  0 [64,17] Array rand ! 'bilinear_mlp.input'
  'b.proj_in' [22,17] Array rand ! 'bilinear_mlp.w.proj_in' ftt
  'b.proj_out' [13,11] Array rand ! 'bilinear_mlp.w.proj_out' ftt
  """
        ),
    ]
    for circ in circs:
        print("Circ")
        circ.print()
        batched = batch_to_concat(circ, 0, 2)
        print(batched)
        batched.print()


def test_batch_einsum():
    settings = OptimizationSettings(max_memory=4_000_000, max_single_tensor_memory=400_000)
    circ = P(
        """
0 Einsum ab,bc,bc->
  1 [1000,1000] Scalar 1.0
  2 [1000,1000] Scalar 1.1
  3 [1000,1000] Scalar 1.2"""
    )
    batched = batch_einsum(circ.cast_einsum(), settings)
    batched.print()


if __name__ == "__main__":
    test_named_axes()
    # test_batching_top()
    # test_batch_einsum()



# %% [markdown]
# **Handcrafted parenthesis balancer and the cumulants approach**
#
# For background on the parenthesis balancing task (and interpretability on it for a DIFFERENT model) see Nix and Fabien's presentation.
#
# It's possible to build a 1L transformer (one attention head, one non-linearity), with hand-selected weights, that completes the parenthesis balancing task perfectly.
#
# This notebook builds such a model and then uses the cumulants approach to attribute its behaviour to different parts of the model.
#
# Two important things we do in this notebook are to use dataset modifications in order to make certain features independent, and also ReLU derivative cumulant approximations.
#
# Skip the next two cells: the first is imports, the second uses a different model to extract the tokens_var. The dataset used keeps the minimum elevation test and the total elevation test independent (look at get_independent_dataset if really curious about the dataset).
#
# Also ensure you have rust and have rust_circuit as in interp/circuit/rust_circuit/readme.md
# %%
# Uncomment to have module autoreload (grr mypy)
# %load_ext autoreload
# %autoreload 2
import os
from typing import Callable, Optional, Tuple

import jax
import numpy as np
import seaborn as sns
import torch
import torch as t

from adversarial.simple_task.dataset_utils import load_dataset
from adversarial.simple_task.model_architectures.simple_transformer import SimpleTokenizer
from interp.circuit import computational_node
from interp.circuit.algebric_rewrite import MUL_REST, rearrange_muls, residual_rewrite
from interp.circuit.circuit import Circuit
from interp.circuit.circuit_model_rewrites import basic_cum_expand_run
from interp.circuit.circuit_utils import cast_circuit
from interp.circuit.computational_node import Add, Einsum, GeneralFunction, Index
from interp.circuit.constant import ArrayConstant, One, Zero
from interp.circuit.cum_algo import cumulant_function_derivative_estim
from interp.circuit.cumulant import Cumulant
from interp.circuit.function_rewrites import get_relu_fake_derivatives
from interp.circuit.get_update_node import FunctionIterativeNodeMatcher as F
from interp.circuit.get_update_node import NameMatcher as NM
from interp.circuit.get_update_node import NodeUpdater as NU
from interp.circuit.get_update_node import Replace
from interp.circuit.print_circuit import PrintCircuit
from interp.circuit.projects.estim_helper import *
from interp.circuit.projects.interp_utils import *
from interp.circuit.projects.punct.utils import standard_pdi
from interp.circuit.scope_manager import ScopeManager
from interp.circuit.scope_rewrites import basic_factor_distribute
from interp.circuit.var import DiscreteVar
from interp.tools.indexer import TORCH_INDEXER as I

sns.set_theme()

os.environ["XLA_PYTHON_CLIENT_ALLOCATOR"] = "platform"
os.environ["XLA_FLAGS"] = "--xla_force_host_platform_device_count=8"  # Use 8 CPU devices
os.environ["RR_CIRCUITS_REPR_NAME"] = "true"

RRFS_DIR = os.path.expanduser("~/rrfs")
RRFS_INTERP_MODELS_DIR = f"{RRFS_DIR}/interpretability_models_jax/"
os.environ["INTERPRETABILITY_MODELS_DIR"] = os.environ.get(
    "INTERPRETABILITY_MODELS_DIR",
    os.path.expanduser("~/interp_models_jax/")
    if os.path.exists(os.path.expanduser("~/interp_models_jax/"))
    else RRFS_INTERP_MODELS_DIR,
)
jax.config.update("jax_platform_name", "cpu")

# %%
task = "balanced_parens"
tokenizer = SimpleTokenizer.task_tokenizer(task)
DEVICE = "cuda" if torch.cuda.is_available() else "cpu"


class ParenBalanceDataSet:
    def __init__(self, data_list, tokenizer):
        self.vocab_size = len(tokenizer.i_to_t)

        strs, is_balanced = [s for s, a in data_list], [a for s, a in data_list]
        self.seqs = np.array(tokenizer.tokenize(strs))
        self.is_balanced = np.array(is_balanced)
        self.strs = strs

    def get_one_hots(self):
        return torch.nn.functional.one_hot(torch.LongTensor(self.seqs), self.vocab_size)


def get_random_paren_list(
    name: str, selector: Callable[[str, bool], bool] = lambda _1, _2: True
) -> Tuple[list[Tuple[str, bool]], list[Tuple[str, bool]]]:
    ds_train = [(s, r) for s, r in load_dataset(task, name)["train"] if selector(s, r)]
    ds_dev = [(s, r) for s, r in load_dataset(task, name)["dev"] if selector(s, r)]
    return ds_train, ds_dev


def all_quite_close(x, y):
    m = torch.max(torch.abs(x.detach().clone().cpu() - y.detach().clone().cpu()))
    return m.item() < 1e-4


def inverse_cumsum(a: torch.Tensor, axis=-1) -> torch.Tensor:
    return torch.flip(torch.flip(a, dims=[axis]).cumsum(dim=axis), dims=[axis])


def count_open_propotions(toks: torch.Tensor) -> torch.Tensor:
    device = toks.device
    if len(list(toks.shape)) == 0:
        return torch.FloatTensor([0.0]).to(device)
    open_parens_counts = inverse_cumsum(toks[..., 3])
    close_parens_counts = inverse_cumsum(toks[..., 4])
    return open_parens_counts / torch.maximum(
        open_parens_counts + close_parens_counts, torch.FloatTensor([1.0]).to(device)
    )


def feature_to_overall_elevation_wrong(x: torch.Tensor, bal_value: int = 0, unbal_value: int = 1) -> torch.Tensor:
    return (unbal_value - bal_value) * x[..., 1] + bal_value


def feature_to_neg_elevation(x: torch.Tensor, bal_value: int = 0, unbal_value: int = 1) -> torch.Tensor:
    return (unbal_value - bal_value) * torch.max(x, dim=-1).values + bal_value


def is_unbalanced(toks: torch.Tensor, func=torch.sigmoid) -> torch.Tensor:
    """
    Take a shape
    (b *) seq_length * vocab_size
    tensor (batch is optional) and calculate whether it is unbalanced (0) or not (1)
    """

    if len(list(toks.shape)) == 0:
        return torch.zeros(1, 1, 1)

    p = count_open_propotions(toks)
    x = feature_to_overall_elevation_wrong(torch.where(torch.abs(p - 0.5) > 0.01, 1.0, 0.0), -10, 20)
    y = feature_to_neg_elevation(torch.where(p - 0.5 > 0.01, 1.0, 0.0), -10, 20)
    return func(x + y)


def total_elev(toks: torch.Tensor) -> torch.Tensor:
    """
    Same as is_unbalanced, except measures whether something fails the total elevation test (1.0) or not (0.0)
    """

    if len(list(toks.shape)) == 0:
        return torch.zeros(1, 1, 1)

    p = count_open_propotions(toks)
    x = feature_to_overall_elevation_wrong(torch.where(torch.abs(p - 0.5) > 0.01, 1.0, 0.0), -10, 10)
    return torch.sigmoid(x)


def min_elev(toks: torch.Tensor, func=torch.sigmoid) -> torch.Tensor:
    """
    Same as is_unbalanced, except measures whether something fails the minimum elevation test (1.0) or not (0.0)
    """

    if len(list(toks.shape)) == 0:
        return torch.zeros(1, 1, 1)

    p = count_open_propotions(toks)
    y = feature_to_neg_elevation(torch.where(p - 0.5 > 0.01, 1.0, 0.0), -10, 10)
    return func(y)


def get_independent_dataset(
    tokenizer=tokenizer,
) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
    """
    Return `INDPENDENT_MIX`: a dataset of
    length 40 parenthesis strings where failing the total elevation test is independent of failing the minimum elevation test. Note that three times as many data points are unbalanced strings compared to balanced strings
    """

    def string_is_good(s: str):
        return len(s) == 40 and s[0] == "("

    random_ds_name = "random_choice_len_40_extra_yeses_16"
    balanced_ds_name = "random_choice_len_40_balanced_advexes"

    unbalanced_selector = lambda s, r: not r and string_is_good(s)
    balanced_selector = lambda s, r: r and string_is_good(s)

    ds_unbalanced_train, _ = get_random_paren_list(random_ds_name, unbalanced_selector)
    ds_balanced_train, _ = get_random_paren_list(balanced_ds_name, balanced_selector)
    big_ds = ParenBalanceDataSet(ds_unbalanced_train, tokenizer)

    m = big_ds.get_one_hots().cpu()

    total_elev_failures = []
    min_elev_failures = []
    both_failures = []
    for i in range(m.shape[0]):
        if not all_quite_close(total_elev(m[i : i + 1]), min_elev(m[i : i + 1])):
            if all_quite_close(total_elev(m[i : i + 1]), torch.zeros(1)):
                min_elev_failures.append(m[i : i + 1])
            else:
                total_elev_failures.append(m[i : i + 1])
        else:
            both_failures.append(m[i : i + 1])

    total_fail_tens = torch.concat(total_elev_failures)
    min_fail_tens = torch.concat(min_elev_failures)
    both_fail_tens = torch.concat(both_failures)

    number_of_balance_to_load = max(len(min_elev_failures), len(total_elev_failures))
    balanced_sample_ds = ParenBalanceDataSet(ds_balanced_train[:number_of_balance_to_load], tokenizer)
    portion_sizes = min(len(min_elev_failures), len(total_elev_failures), len(both_failures))
    return (
        torch.cat(
            (
                both_fail_tens[:portion_sizes],
                total_fail_tens[:portion_sizes],
                min_fail_tens[:portion_sizes],
                balanced_sample_ds.get_one_hots()[:portion_sizes],
            ),
            0,
        ),
        None,
    )


def get_sub_circuit(circuit: Circuit, child_name: str):
    """
    Get a circuit leaf in the computational graph of `circuit`
    that has name `child_name`
    """

    assert isinstance(circuit, Circuit)
    assert isinstance(child_name, str)
    return F(NM(child_name)).g().get_unique_c(circuit)


def get_truth_for_cumulant(model: Circuit, function, name: str = "true") -> Circuit:
    """
    Return the circuit representing the input parenthesis sequence
    """

    tokens = get_sub_circuit(model, "tokens")

    return Index(
        GeneralFunction(
            tokens,
            lambda x: torch.broadcast_to(function(x)[:, None, None], x.shape),
            non_batch_dims=(-1, -2),
            name=f"{name}_function",
        ),
        I[0, 0],
        name=name,
    )


def replace_tokens(data: torch.Tensor, *circuits: Circuit):
    """
    Replace the tokens inside all circuits
    """

    new_tokens = ArrayConstant.from_converted(data, dtype=torch.float, name="tokens", device=DEVICE)

    new_node = DiscreteVar(new_tokens, name="tokens_var")

    return *[NU(Replace(new_node), F(NM("tokens")))(c) for c in circuits], new_node


do_cumulant_evaluations = True
sample_ds, val_ds = get_independent_dataset(tokenizer=tokenizer)
#%%
def make_handcrafted_transformer():
    """
    `tokens_var` is 42 * 5 in shape
    """
    # Embedding
    # We encode ( (3) as e_1, and ) (4) as e_2 in the 2D residual stream
    # also, encode the end of sequence token as 1 (this is needed for the total elevation test)
    m = torch.zeros(5, 2)
    m[3][0] = 1.0
    m[4][1] = 1.0
    m[0][0] = 1.0

    fake_tokens_var = Zero(shape=(42, 5), name="tokens")

    embed_matrix = ArrayConstant(m, name="w.embed")
    embedded = Einsum.from_einsum_str("lo,ox->lx", fake_tokens_var, embed_matrix, name="embed")

    # QKV weights
    # we make queries for both e_1 and e_2 be 10.0 (queries and keys are 1D)
    # on the other hand the value vectors are +1 for open brackets and -1 for close brackets
    qw = ArrayConstant(torch.ones(2, 1) * 10, name="q.w")
    kw = One(shape=(2, 1), name="k.w")
    m2 = t.zeros(2, 1)
    m2[0][0] = 1.0
    m2[1][0] = -1.0
    vw = ArrayConstant(m2, name="v.w")

    # QKV calculations
    # note we also use masking, so everything only pays attention to things before it
    q = Einsum.from_einsum_str("io,li->lo", qw, embedded, name="q")
    k = Einsum.from_einsum_str("io,li->lo", kw, embedded, name="k")
    v = Einsum.from_einsum_str("io,li->lo", vw, embedded, name="v")
    qk = Einsum.from_einsum_str("lo,Lo->lL", q, k, name="qk")
    upper_triangular_tensor = t.ones(42, 42).float() * (-1e6)
    for i in range(42):
        for j in range(42):
            if i <= j:
                upper_triangular_tensor[i][j] = 0
    upper_triangular = ArrayConstant(upper_triangular_tensor, name="upper_triangular")
    scores = Add.from_unweighted_list([qk, upper_triangular], name="masked")
    probs = computational_node.softmax(scores, name="probs")
    values = Einsum.from_einsum_str("lL,Lo->lo", probs, v, name="head.out")

    # this setup means (as you can verify!) if any of the last 39 positions are positive, the elevation test has failed. Additionally if the last position (which attend to all the sequence) is not <= 0, the string is unbalanced,
    # So we need reverse this positions sign, then take a ReLU and if ANYWHERE is positive then the string is unbalanced; this is pretty much an MLP!
    m3 = t.eye(42).float()
    m3[0, 0] = -1.0  # I thonk...
    m3[:, 41] = 0
    m3[41, :] = 0
    m1w = ArrayConstant(m3, name="m1.w")
    m1ed = Einsum.from_einsum_str("lL,lo->L", m1w, values, name="m1.out")
    relued = computational_node.relu(m1ed, name="m1.act")

    # then add up all the entries
    m2w = ArrayConstant(t.ones(42, 42), name="m2.w")
    m2ed = Einsum.from_einsum_str("lL,l->L", m2w, relued, name="m2.out")

    # taking sigmoid and the convention that 0.5 rounds down to 0, this model achieves perfect performance on length 40 sequences
    logits = Index(m2ed, I[0], name="logits")
    return logits


circuit = make_handcrafted_transformer()
truth = get_truth_for_cumulant(circuit, is_unbalanced)
circuit = cast_circuit(circuit, device=DEVICE)
truth = cast_circuit(truth, device=DEVICE)
circuit, truth, new_node = replace_tokens(sample_ds, circuit, truth)
# %%
# some tools for printing circuits
d1 = PrintCircuit(
    print_html=True, colorize=lambda _: True, max_depth=1, print_shape=True, copy_button=True
)  # depth 1 printing, but collapsibles mean things can be expanded

print("View the collapsible tree for the circuit:")
d1(circuit)
# %%
# Now let's define the covariance we want to explain
# (why is the covariance between the true answer and the difference between the two output logits high?)

cumulant_circuit = Cumulant((circuit, truth))
scopes = [ScopeManager(cumulant_circuit)]  # all scopes considered, as a list because checking how things change
d1(scopes[-1])
#%% [markdown]
# We now need to "push down" the cumulant. This is equivalent to noticing that the output logits $L$ are equal to (an entry in) $Mx$, where $M$ is the final matrix multiply.
# So we have $K(Mx, T) = MK(x, T)$, as $M$ is constant. So "pushing down" the cumulant means factoring out the $M$ term:

# %%
scopes.append(
    scopes[-1].u(
        NU(
            lambda x: basic_cum_expand_run(  # expand cumulant
                Cumulant.unwrap(x),
                cum_expand_through_factored_add_concat=True,
                until_func=NM("true"),  # do not expand through "true"
                child_suffix="logits",  # ... instead expand through "logits"
            ),
            F(NM(scopes[-1].unique().name)),  # the name of the cumulant that we want to expand
        )
    )
)
d1(scopes[-1])

#%%
print("Now push down the index in a similar way:")
scopes.append(scopes[-1].u(NU(lambda x: standard_pdi(x), F(NM(scopes[-1].unique().name)))))
# upto here, maybe delete ya extra type safety
#%%
print("Now let's see what the covariance is!")
estim = EstimHelper(Zero(), use_new_estim=True, device=DEVICE)
list(estim.estimate(scopes[-1].unique()).items())[0][1].item()
#%%
print("Now let's expand the derivative non-linearity by estimating with the first order term")


def relu_rewrite(
    scope: ScopeManager, cumulant_name: str, relu_name: str, highest_deriv: int = 6, std_deviation: float = 0.1
):
    """
    Expand a ReLU non-linearity (GeneralFunction) `relu_name`, which has a parent (Cumulant) `cumulant_name, into `highest_deriv` number of terms. Approximates ReLUs with the derivative of a normal PDF with std `std_deviation`
    """

    def run_expand_deriv_relu(x: Circuit, relu_name: str):
        """
        Function that updates ReLU nodes
        """

        pre_act = GeneralFunction.unwrap(F(NM(relu_name)).g().get_unique_c(x))

        estim = cumulant_function_derivative_estim(
            Cumulant.unwrap(x),
            pre_act,
            highest_deriv=highest_deriv,
            get_derivative=get_relu_fake_derivatives(pre_act, fake_dirac_std=std_deviation),
        )
        out, _, _ = residual_rewrite(x, estim, running_name="d_estim")
        return out

    return scope.u(NU(lambda x: run_expand_deriv_relu(x, relu_name), F(NM(cumulant_name))), no_eq_besides_name=False)


scopes.append(relu_rewrite(scopes[-1], cumulant_name="I k2 m1.act, true I", relu_name="m1.act"))
d1(scopes[-1])
#%%
scopes.append(basic_factor_distribute(scopes[-1], modify_inplace=False))
#%% The tree looks quite different now!
# Note the Jacobians here (which are large and mean we won't be able to evaluate many of these cumulants) are actually diagonal, as ReLU acts elementwise. See batched_deriv for a more efficient use (which we don't need here as we only need lower order terms)
scopes.append(
    scopes[-1].sub_get(F(NM("I k2 m1.act, true I_d_estim_out")))
)  # get derivative terms (we're now approximating the original cumulant)
#%%
scopes.append(
    scopes[-1].u(
        NU(
            lambda x: basic_cum_expand_run(  # expand cumulant
                Cumulant.unwrap(x),
                cum_expand_through_factored_add_concat=True,
                until_func=NM("true"),
                child_suffix="m1.out",
            ),
            F(NM("I k2 m1.out, true I")),
        ),
        no_eq_besides_name=False,  # now needed...
    )
)
#%%
scopes.append(basic_factor_distribute(scopes[-1], modify_inplace=False))
#%%
scopes.append(
    scopes[-1].sub_get(F(NM("I k2 m1.act, true I_d_estim_item_1_out")))
)  # now it is a scalar we get the first derivative
#%%
print("Let's check the cumulant between this approximation and the truth:")
print(list(estim.estimate(scopes[-1].unique()).items())[0][1].item())
print(
    "Note that this is less than half of the total effect size. It turns out it is possible to explain behaviour in this case though"
)
#%%
scopes.append(
    scopes[-1].u(
        NU(
            lambda x: basic_cum_expand_run(  # expand cumulant
                Cumulant.unwrap(x),
                cum_expand_through_factored_add_concat=True,
                until_func=NM("true"),
                child_suffix="head.out",
            ),
            F(NM("I k2 head.out, true I")),
        ),
        no_eq_besides_name=False,  # now needed...
    )
)
#%%
scopes.append(basic_factor_distribute(scopes[-1], modify_inplace=False))
#%%
circuits_to_estimate = Add.unwrap(scopes[-1].unique()).items.keys()
out = estim.estimate(*circuits_to_estimate)
for k, v in out.items():
    print(k.name, v.item())
print("This is not surprising since recall that the attention pattern was constant")
#%%
scopes.append(scopes[-1].sub_get(F(NM("I k2 true, v k1 probs I perm_out"))))
#%%
print("Finally we can rearrage the Einsums to get a product of the corellatios with the input, and another matrix:")
final_circuit = rearrange_muls(Einsum.unwrap(scopes[-1].unique()), (0, MUL_REST))
d1(final_circuit)

#%%
q = list(estim.estimate(get_sub_circuit(final_circuit, "I k2 tokens_var, true I")).items())[0][1][:, 3:]


def plot_matrix(m):
    if len(m.shape) < 2:
        m.unsqueeze(0)
    sns.heatmap(m.detach().cpu())


plot_matrix(q)
print("This basically shows us correlations between the input and the truth, so doesn't give much signal")
# %%
q2 = list(
    estim.estimate(
        get_sub_circuit(final_circuit, "m1.w m1.act_deriv_d_estim_expectation d_estim_recip_1! m2.w_idxed")
    ).items()
)[0][
    1
]  # [:, 3:]
plot_matrix(q2.unsqueeze(0))
print(
    "The high value at 0 is meaningless since it is timesed by 0 in the dataset corellation term. The rest of the plot shows that the final position accounts for most of the variance (the total elevation test) and the cyclical bands show the minimum elevation test occurs from right to left as the first minimum elevation failure will always be at an odd time"
)



from typing import Any, Dict, Optional, cast

import einops
import torch
from attrs import frozen

import rust_circuit as rc
import rust_circuit.module_library as mod_l
import rust_circuit.optional as op
from interp.circuit.circuit import Circuit, MemoizedFn
from interp.circuit.circuit_models import attention, batchnorm, bilinear_mlp, layernorm, pointwise_nonlin_mlp
from interp.circuit.circuit_utils import evaluate_fn
from interp.circuit.constant import ArrayConstant
from interp.model.blocks import BatchNormFixed, LayerNorm
from interp.model.gpt_modules import Attention, GptBlock, MlpActType


def evaluate_circ(c: Circuit) -> torch.Tensor:
    return MemoizedFn(evaluate_fn(dtype=torch.float64))(c)


@frozen
class FakeLinear:
    weights: torch.Tensor
    bias: Optional[torch.Tensor]

    @property
    def use_bias(self) -> bool:
        return self.bias is not None

    def get_weights(self) -> torch.Tensor:
        return self.weights

    def get_bias(self) -> torch.Tensor:
        return op.unwrap(self.bias)


@frozen
class FakeGptBlock:
    linear1: FakeLinear
    linear2: FakeLinear
    mlp_act_type: MlpActType = "bilinear"


def raw_test_bilinear_mlp(hidden: int, out_hidden: int, bilinear_mlp_proj: int, batch: rc.Shape, output_bias: bool):
    proj_in = torch.randn(bilinear_mlp_proj * 2, hidden, dtype=torch.float64)
    in_bias = torch.randn(bilinear_mlp_proj * 2, dtype=torch.float64)
    proj_out = torch.randn(out_hidden, bilinear_mlp_proj, dtype=torch.float64)
    out_bias = torch.randn(out_hidden, dtype=torch.float64)
    inp = torch.randn(*batch, hidden, dtype=torch.float64)
    dummy_proj_out = torch.cat([proj_out, torch.randn(proj_out.shape, dtype=torch.float64)], dim=-1)
    b = FakeGptBlock(FakeLinear(proj_in, in_bias), FakeLinear(dummy_proj_out, out_bias if output_bias else None))
    old_circ = bilinear_mlp(cast(GptBlock, b), ArrayConstant(inp.reshape(1, -1, hidden)))

    mod = rc.Module(
        mod_l.get_bilinear_mlp(output_bias=output_bias).spec,
        **{
            "m.w.proj_in": rc.Array(proj_in),
            "m.w.in_bias": rc.Array(in_bias),
            "m.w.proj_out": rc.Array(proj_out),
            "m.input": rc.Array(inp),
        },
        **({"m.w.out_bias": rc.Array(out_bias)} if output_bias else {}),
        name="",
    )

    torch.testing.assert_close(mod.evaluate(), evaluate_circ(old_circ).reshape(*batch, out_hidden))


def test_bilinear_mlp():
    raw_test_bilinear_mlp(2, 2, 3, (), True)
    raw_test_bilinear_mlp(2, 2, 3, (), False)
    raw_test_bilinear_mlp(2, 5, 3, (), False)
    raw_test_bilinear_mlp(2, 5, 3, (1,), False)
    raw_test_bilinear_mlp(2, 5, 3, (7,), False)
    raw_test_bilinear_mlp(2, 5, 3, (7,), True)
    raw_test_bilinear_mlp(2, 5, 3, (7, 3), True)


def raw_test_pointwise_mlp(
    hidden: int, out_hidden: int, mlp_proj: int, batch: rc.Shape, output_bias: bool, mlp_act_type: MlpActType
):
    proj_in = torch.randn(mlp_proj, hidden, dtype=torch.float64)
    in_bias = torch.randn(mlp_proj, dtype=torch.float64)
    proj_out = torch.randn(out_hidden, mlp_proj, dtype=torch.float64)
    out_bias = torch.randn(out_hidden, dtype=torch.float64)
    inp = torch.randn(*batch, hidden, dtype=torch.float64)

    b = FakeGptBlock(
        FakeLinear(proj_in, in_bias), FakeLinear(proj_out, out_bias if output_bias else None), mlp_act_type=mlp_act_type
    )
    old_circ = pointwise_nonlin_mlp(cast(GptBlock, b), ArrayConstant(inp.reshape(1, -1, hidden)))

    mod = rc.Module(
        mod_l.get_pointwise_mlp(function_str=mlp_act_type, output_bias=output_bias).spec,
        **{
            "m.w.proj_in": rc.Array(proj_in),
            "m.w.in_bias": rc.Array(in_bias),
            "m.w.proj_out": rc.Array(proj_out),
            "m.input": rc.Array(inp),
        },
        **({"m.w.out_bias": rc.Array(out_bias)} if output_bias else {}),
        name="",
    )

    torch.testing.assert_close(mod.evaluate(), evaluate_circ(old_circ).reshape(*batch, out_hidden))


def test_pointwise_mlp():
    raw_test_pointwise_mlp(2, 2, 3, (), True, "gelu")
    raw_test_pointwise_mlp(2, 2, 3, (), False, "gelu")
    raw_test_pointwise_mlp(2, 5, 3, (), False, "gelu")
    raw_test_pointwise_mlp(2, 5, 3, (1,), False, "gelu")
    raw_test_pointwise_mlp(2, 5, 3, (7,), False, "gelu")
    raw_test_pointwise_mlp(2, 5, 3, (7,), True, "gelu")
    raw_test_pointwise_mlp(2, 5, 3, (7, 3), True, "gelu")
    raw_test_pointwise_mlp(2, 2, 3, (), True, "relu")
    raw_test_pointwise_mlp(2, 2, 3, (), False, "relu")
    raw_test_pointwise_mlp(2, 5, 3, (7, 3), True, "relu")


@frozen
class FakeAttention:
    q: torch.Tensor
    k: torch.Tensor
    v: torch.Tensor
    o: torch.Tensor
    q_bias: Optional[torch.Tensor]
    k_bias: Optional[torch.Tensor]
    v_bias: Optional[torch.Tensor]
    o_bias: Optional[torch.Tensor]

    @property
    def bias(self) -> bool:
        return self.q_bias is not None

    @property
    def hidden_size(self) -> int:
        return self.q.shape[-1]

    @property
    def num_heads(self) -> int:
        return self.q.shape[0]

    def get_qkv_mats(self):
        return self.q, self.k, self.v

    @property
    def attn_weights(self) -> FakeLinear:
        bias: Optional[torch.Tensor] = None
        if self.q_bias is not None:
            bias = torch.cat(
                [op.unwrap(self.q_bias).flatten(), op.unwrap(self.k_bias).flatten(), op.unwrap(self.v_bias).flatten()],
                dim=0,
            )
        return FakeLinear(
            torch.cat(
                [
                    self.q.reshape(-1, self.hidden_size),
                    self.k.reshape(-1, self.hidden_size),
                    self.v.reshape(-1, self.hidden_size),
                ],
                dim=0,
            ),
            bias,
        )

    @property
    def project_output(self) -> FakeLinear:
        return FakeLinear(einops.rearrange(self.o, "num_heads out v -> out (num_heads v)"), self.o_bias)

    @property
    def causal_mask(self) -> bool:
        return False

    @property
    def dtype(self):
        return torch.float64


def raw_test_attention(
    hidden: int,
    qkv_dim: int,
    num_heads: int,
    out_hidden: int,
    seq_len: int,
    batch: rc.Shape,
    use_bias: bool,
    use_pos: bool,
):
    qk_dim = v_dim = qkv_dim  # coverage isn't great because we can't test different shape case
    q = torch.randn(num_heads, qk_dim, hidden, dtype=torch.float64)
    q_bias = torch.randn(num_heads, qk_dim, dtype=torch.float64)
    k = torch.randn(num_heads, qk_dim, hidden, dtype=torch.float64)
    k_bias = torch.randn(num_heads, qk_dim, dtype=torch.float64)
    v = torch.randn(num_heads, v_dim, hidden, dtype=torch.float64)
    v_bias = torch.randn(num_heads, v_dim, dtype=torch.float64)
    o = torch.randn(num_heads, out_hidden, v_dim, dtype=torch.float64)
    o_bias = torch.randn(out_hidden, dtype=torch.float64)
    inp = torch.randn(*batch, seq_len, hidden, dtype=torch.float64)
    pos_inp = torch.randn(*batch, seq_len, hidden, dtype=torch.float64)

    extra_args: Dict[str, rc.Circuit] = {}

    if not use_bias:
        q_bias_v: Optional[torch.Tensor] = None
        k_bias_v: Optional[torch.Tensor] = None
        v_bias_v: Optional[torch.Tensor] = None
        o_bias_v: Optional[torch.Tensor] = None
    else:
        extra_args = {
            "a.w.q_bias": rc.Array(q_bias * (2 if use_pos else 1)),
            "a.w.k_bias": rc.Array(k_bias * (2 if use_pos else 1)),
            "a.w.v_bias": rc.Array(v_bias),
            "a.w.o_bias": rc.Array(o_bias),
        }

        q_bias_v = q_bias
        k_bias_v = k_bias
        v_bias_v = v_bias
        o_bias_v = o_bias

    if use_pos:
        extra_args["a.pos_input"] = rc.Array(pos_inp)

    mask = torch.bernoulli(torch.full((*batch, seq_len, seq_len), 0.5)).double()

    old_circ = attention(
        cast(Attention, FakeAttention(q, k, v, o, q_bias_v, k_bias_v, v_bias_v, o_bias_v)),
        ArrayConstant(inp.reshape(-1, seq_len, hidden)),
        position_embedding=ArrayConstant(pos_inp.reshape(-1, seq_len, hidden)) if use_pos else None,
        # for some reason, this is negated relative to what you would think...
        pos_mask=ArrayConstant(1.0 - mask.reshape(-1, seq_len, seq_len), name="inp_mask"),
    )

    args: dict[str, rc.Circuit] = {
        "a.w.q": rc.Array(q),
        "a.w.k": rc.Array(k),
        "a.w.v": rc.Array(v),
        "a.w.o": rc.Array(o),
        "a.mask": rc.Array(mask),
        "a.input": rc.Array(inp),
        **extra_args,
    }
    args = {s: c.rename("arr." + s) for s, c in args.items()}

    mod = rc.Module(
        mod_l.get_attention(bias=use_bias, pos=use_pos).spec,
        **args,
        name="",
    )

    torch.testing.assert_close(mod.evaluate(), evaluate_circ(old_circ).reshape(*batch, seq_len, out_hidden))


def test_attention():
    torch.manual_seed(3854)
    raw_test_attention(1, 1, 1, 1, 2, (), use_bias=False, use_pos=False)
    raw_test_attention(1, 1, 2, 1, 3, (), use_bias=False, use_pos=False)
    raw_test_attention(2, 3, 5, 7, 11, (), use_bias=False, use_pos=False)
    raw_test_attention(2, 3, 5, 7, 11, (1,), use_bias=False, use_pos=False)
    raw_test_attention(2, 3, 5, 7, 11, (3,), use_bias=False, use_pos=False)
    raw_test_attention(2, 3, 5, 7, 11, (4, 3, 2), use_bias=False, use_pos=False)
    raw_test_attention(2, 3, 5, 7, 11, (3,), use_bias=True, use_pos=False)
    raw_test_attention(2, 3, 5, 7, 11, (3,), use_bias=False, use_pos=True)
    raw_test_attention(2, 3, 5, 7, 11, (3,), use_bias=True, use_pos=True)
    raw_test_attention(2, 3, 5, 7, 11, (), use_bias=True, use_pos=True)
    raw_test_attention(2, 3, 5, 7, 11, (4, 3), use_bias=True, use_pos=True)


@frozen
class FakeNorm:
    variables: Dict[str, Any]
    epsilon: float


def raw_test_layernorm(hidden: int, batch: rc.Shape):
    scale = torch.randn(hidden, dtype=torch.float64)
    bias = torch.randn(hidden, dtype=torch.float64)
    inp = torch.randn(*batch, hidden, dtype=torch.float64)

    variables = {"params": {"scale": scale, "bias": bias}}
    norm_b = FakeNorm(variables, epsilon=1e-5)
    old_circ = layernorm(cast(LayerNorm, norm_b), ArrayConstant(inp.reshape(1, -1, hidden)))

    mod = rc.Module(
        mod_l.layernorm.spec,
        **{"ln.w.bias": rc.Array(bias), "ln.w.scale": rc.Array(scale), "ln.input": rc.Array(inp)},
        name="",
    )

    torch.testing.assert_close(mod.evaluate(), evaluate_circ(old_circ).reshape(*batch, hidden))


def test_layernorm():
    raw_test_layernorm(2, ())
    raw_test_layernorm(2, (3, 4))
    raw_test_layernorm(2, (3, 4))
    raw_test_layernorm(2, (3, 4, 7))
    raw_test_layernorm(9, (3, 4, 7))


def raw_test_batchnorm(hidden: int, batch: rc.Shape):
    scale = torch.randn(hidden, dtype=torch.float64)
    bias = torch.randn(hidden, dtype=torch.float64)
    mean = torch.randn(hidden, dtype=torch.float64)
    var = torch.randn(hidden, dtype=torch.float64) ** 2
    inp = torch.randn(*batch, hidden, dtype=torch.float64)

    variables = dict(params=dict(scale=scale, bias=bias, mean=mean, var=var))
    norm_b = FakeNorm(variables, epsilon=1e-5)
    old_circ = batchnorm(cast(BatchNormFixed, norm_b), ArrayConstant(inp.reshape(1, -1, hidden)))

    mod = rc.Module(
        mod_l.batchnorm.spec,
        **{
            "bn.w.bias": rc.Array(bias),
            "bn.w.scale": rc.Array(scale),
            "bn.var": rc.Array(var),
            "bn.mean": rc.Array(mean),
            "bn.input": rc.Array(inp),
        },
        name="",
    )

    torch.testing.assert_close(mod.evaluate(), evaluate_circ(old_circ).reshape(*batch, hidden))


def test_batchnorm():
    raw_test_batchnorm(2, ())
    raw_test_batchnorm(2, (3, 4))
    raw_test_batchnorm(2, (3, 4))
    raw_test_batchnorm(2, (3, 4, 7))
    raw_test_batchnorm(9, (3, 4, 7))


def raw_test_nll(classes: int, batch: rc.Shape):
    inp = torch.randn(*batch, classes)
    labels = torch.randint(0, classes, batch).long()
    expected = torch.nn.functional.cross_entropy(
        inp.reshape(-1, classes), labels.flatten().long(), reduction="none"
    ).reshape(batch)
    actual = rc.Module(
        mod_l.negative_log_likelyhood.spec,
        **{
            "ll.input": rc.Array(inp),
            "ll.label": rc.Array(labels),
        },
        name="",
    )

    torch.testing.assert_close(actual.evaluate(), expected)


def test_nll():
    raw_test_nll(1, ())
    raw_test_nll(2, ())
    raw_test_nll(2, (3, 4))
    raw_test_nll(2, (3, 4))
    raw_test_nll(2, (3, 4, 7))
    raw_test_nll(9, (3, 4, 7))
    raw_test_nll(1, (3, 4, 7))



from rust_circuit import Array
from rust_circuit.py_utils import timed


def test_serialization():
    shape = (1024, 10)
    with timed("init"):
        arr = Array.randn(*shape)
    with timed("save"):
        arr.save_rrfs()
    hash_base16 = arr.tensor_hash_base16()
    print(hash_base16)
    with timed("load"):
        arr2 = Array.from_hash(None, hash_base16)
    print(arr == arr2)
    arr.print()
    arr2.print()


if __name__ == "__main__":
    test_serialization()
    print("done")



import random
from typing import Iterable, Optional, Tuple

import einops
import pytest
import torch

import rust_circuit as rc
from rust_circuit import (
    Add,
    Array,
    Circuit,
    Concat,
    Cumulant,
    DiscreteVar,
    Einsum,
    Parser,
    RandomSampleSpec,
    Rearrange,
    RunDiscreteVarAllSpec,
    Sampler,
    Scalar,
    Shape,
    Symbol,
    Tag,
    default_hash_seeder,
    factored_cumulant_expectation_rewrite,
    multinomial,
    optimize_and_evaluate,
)
from rust_circuit.py_utils import I


def test_estim_cumulant_shape_correct(seed=23473):
    random.seed(seed)

    items = [
        Symbol.new_with_random_uuid(shape=()),
        Symbol.new_with_random_uuid(shape=(2,)),
        Symbol.new_with_random_uuid(shape=(3,)),
        Symbol.new_with_random_uuid(shape=(2, 3)),
        Symbol.new_with_random_uuid(shape=(4, 5)),
        Symbol.new_with_random_uuid(shape=(4, 5, 6, 7)),
        Symbol.new_with_random_uuid(shape=(8, 9, 10, 11, 12, 13, 17)),
        Symbol.new_with_random_uuid(shape=(18, 19, 20)),
    ]

    for with_var in [False, True]:
        if with_var:
            n_vals = 7
            probs = DiscreteVar.uniform_probs_and_group(n_vals)
            new_items = [DiscreteVar(x.expand_at_axes(0, n_vals), probs) for x in items]
            all_probs = [probs]
        else:
            all_probs = []
            new_items = items

        for count in range(7):
            for _ in range(5):
                random.shuffle(new_items)
                cumulant = Cumulant(*new_items[:count])
                transformed = Sampler(RunDiscreteVarAllSpec(all_probs)).estimate(cumulant)
                assert transformed.shape == cumulant.shape


def test_correct_cumulant_estimate(seed=283838):
    torch.manual_seed(seed)
    x = torch.randn(5, 3, 2, dtype=torch.float64)
    y = torch.nn.functional.softmax(x, dim=-1).sum(dim=1)
    z = x ** 3
    w = torch.sigmoid(x[:, :, 0])

    vals = [x, y, z, w]

    var_x = DiscreteVar.new_uniform(Array(x))
    vars_unweighted = [DiscreteVar(Array(v), var_x.probs_and_group) for v in vals]

    probs = torch.rand(5, dtype=torch.float64)
    probs = probs / probs.sum()

    probs_and_group = Array(probs)

    vars_weighted = [DiscreteVar(Array(v), probs_and_group) for v in vals]

    def estimate_and_comp(circuit, actual):
        # todo: add is_full to rust
        transformed = (
            Sampler(
                RunDiscreteVarAllSpec.create_full_from_circuits(circuit),
            )
            .estimate(circuit)
            .evaluate()
        )
        torch.testing.assert_close(
            transformed,
            actual,
            atol=1e-6,
            rtol=1e-6,
        )

    for all_vars, weights in [
        [vars_unweighted, torch.full((5,), 1 / 5, dtype=torch.float64)],
        [vars_weighted, probs],
    ]:
        var_x, var_y, var_z, var_w = all_vars
        all_means = [torch.einsum("i ..., i -> ...", v, weights).unsqueeze(0) for v in vals]
        c_x, c_y, c_z, c_w = [v - m for v, m in zip(vals, all_means)]
        x_mean = all_means[0].squeeze(0)
        estimate_and_comp(
            Cumulant(
                var_x,
            ),
            x_mean,
        )

        uncorrected_cov = torch.einsum("b i j, b k, b -> i j k", c_x, c_y, weights)
        estimate_and_comp(Cumulant(var_x, var_y), uncorrected_cov)

        uncorrected_third_cum = torch.einsum("b i j, b k, b l m, b -> i j k l m", c_x, c_y, c_z, weights)
        estimate_and_comp(Cumulant(var_x, var_y, var_z), uncorrected_third_cum)

        centered_4th_mom = torch.einsum("b i j, b k, b l m, b o, b -> i j k l m o", c_x, c_y, c_z, c_w, weights)
        first_pair = torch.einsum(
            "i j k, l m o -> i j k l m o",
            torch.einsum("b i j, b k, b -> i j k", c_x, c_y, weights),
            torch.einsum("b l m, b o, b -> l m o", c_z, c_w, weights),
        )
        snd_pair = torch.einsum(
            "i j l m, k o -> i j k l m o",
            torch.einsum("b i j, b l m, b -> i j l m", c_x, c_z, weights),
            torch.einsum("b k, b o, b-> k o", c_y, c_w, weights),
        )
        thrd_pair = torch.einsum(
            "i j o, k l m -> i j k l m o",
            torch.einsum("b i j, b o, b -> i j o", c_x, c_w, weights),
            torch.einsum("b k, b l m, b -> k l m", c_y, c_z, weights),
        )

        pair_sum = first_pair + snd_pair + thrd_pair

        uncorrected_4th_cum = centered_4th_mom - pair_sum

        estimate_and_comp(Cumulant(var_x, var_y, var_z, var_w), uncorrected_4th_cum)


def test_trivial_factored_cumulant_expectation_rewrite():
    assert factored_cumulant_expectation_rewrite(Cumulant()) == Cumulant()
    cum = Cumulant(DiscreteVar.new_uniform(Array.randn(3, 2)))
    assert factored_cumulant_expectation_rewrite(cum) == cum


def raw_test_factored_cumulant_expectation_rewrite_vs_basic_expectation(cumulant: Cumulant):
    transform = Sampler(RunDiscreteVarAllSpec.create_full_from_circuits(cumulant))
    torch.testing.assert_close(
        transform.estimate(factored_cumulant_expectation_rewrite(cumulant)).evaluate(),
        transform.estimate(cumulant).evaluate(),
    )


def test_factored_cumulant_expectation_rewrite_vs_basic_expectation_simple():
    # could add hypothesis test, but feels extra...
    shapes = (2, 3), (4, 5, 1), (2, 1, 4), (3, 2), (1, 3, 1, 1, 1), (2, 4, 1)
    n_vals = 11
    group = DiscreteVar.uniform_probs_and_group(n_vals)
    variables = [DiscreteVar(Array.randn(*((n_vals,) + s)), group) for s in shapes]

    for i in range(2, len(variables) + 1):
        raw_test_factored_cumulant_expectation_rewrite_vs_basic_expectation(Cumulant(*variables[:i]))


def make_discrete_var(
    values_shape: Shape, batch_size: int, name: Optional[str] = None, group: Optional[Circuit] = None
):
    vals = Array.randn(batch_size, *values_shape)
    out = DiscreteVar(vals, group, name)
    return out, vals.value, out.probs_and_group.evaluate()


def make_discrete_var_group(n: int, name: Optional[str] = None):
    x = torch.rand(n)
    x /= x.sum()
    return Tag.new_with_random_uuid(Array(x), name)


def make_discrete_var_test_circuits(
    seed=238832, just_mean: bool = False
) -> Iterable[Tuple[Circuit, torch.Tensor, torch.Tensor, Circuit]]:
    torch.manual_seed(seed)

    x, x_value, x_probs_and_group = make_discrete_var((3, 2), 17, "x")
    yield x, x_value, x_probs_and_group, x.probs_and_group

    y, y_value, _ = make_discrete_var((3, 42), 17, "y", group=x.probs_and_group)
    yield y, y_value, x_probs_and_group, x.probs_and_group

    z, z_value, _ = make_discrete_var((3, 7, 6, 2), 17, "z", group=x.probs_and_group)
    yield z, z_value, x_probs_and_group, x.probs_and_group

    z_rearrange = Rearrange.from_string(z, "a b c d -> d (c b) a", name="z_rearrange")
    z_rearrange_eval = einops.repeat(z_value, "batch a b c d -> batch d (c b) a")
    yield z_rearrange, z_rearrange_eval, x_probs_and_group, x.probs_and_group

    const_l = Array.randn(6, 2, name="const_l")
    const_r = Array.randn(3, 7, name="const_r")

    z_mul_const = Einsum.from_einsum_string("a b c d, c d, a b -> b c", z, const_l, const_r)
    z_mul_const_eval = torch.einsum("e a b c d, c d, a b -> e b c", z_value, const_l.value, const_r.value)
    yield z_mul_const, z_mul_const_eval, x_probs_and_group, x.probs_and_group

    const_single_l = Array.randn(4, name="const_single_r")
    const_single_r = Array.randn(4, name="const_single_r")

    simple_trace = Einsum.from_einsum_string("a,a->aa", const_single_l, const_single_r)
    simple_trace_eval = einops.repeat(
        torch.diag(torch.einsum("a, a -> a", const_single_l.value, const_single_r.value)),
        "a b -> c a b",
        c=x_probs_and_group.shape[0],
    )
    yield simple_trace, simple_trace_eval, x_probs_and_group, x.probs_and_group

    z_repeated_back_to_orig = z_mul_const.expand_at_axes((0, -1), (3, 2))
    z_repeated_back_to_orig_eval = einops.repeat(z_mul_const_eval, "batch b c -> batch a b c d", a=3, d=2)
    yield z_repeated_back_to_orig, z_repeated_back_to_orig_eval, x_probs_and_group, x.probs_and_group

    z_repeated_back_to_orig_other_order = z_mul_const.expand_at_axes((-1, 0), (2, 3))
    yield z_repeated_back_to_orig_other_order, z_repeated_back_to_orig_eval, x_probs_and_group, x.probs_and_group

    z_unsqueeze = z_mul_const.unsqueeze((0, -1))
    z_unsqueeze_eval = einops.repeat(z_mul_const_eval, "batch b c -> batch 1 b c 1")
    yield z_unsqueeze, z_unsqueeze_eval, x_probs_and_group, x.probs_and_group

    z_mul_const_rearrange = Rearrange.from_string(
        z_repeated_back_to_orig, "a b c d -> d (c b) a", name="z_mul_const_rearrange"
    )
    z_mul_const_rearrange_eval = einops.repeat(z_repeated_back_to_orig_eval, "batch a b c d -> batch d (c b) a")
    yield z_mul_const_rearrange, z_mul_const_rearrange_eval, x_probs_and_group, x.probs_and_group

    y_dup, y_dup_value, _ = make_discrete_var((3, 42), 17, "y_dup", group=x.probs_and_group)
    yield y_dup, y_dup_value, x_probs_and_group, x.probs_and_group

    y_sums = y.add(y_dup)
    y_sums_value = y_value + y_dup_value
    yield y_sums, y_sums_value, x_probs_and_group, x.probs_and_group

    new_group = make_discrete_var_group(7, name="non_uniform_group_and_probs")
    with_variable_weights, with_variable_weights_value, with_variable_weights_group = make_discrete_var(
        (2, 3), 7, "var_weights", group=new_group
    )
    yield (
        with_variable_weights,
        with_variable_weights_value,
        with_variable_weights_group,
        with_variable_weights.probs_and_group,
    )

    with_variable_weights_other, with_variable_weights_other_val, _ = make_discrete_var(
        (2,), 7, "var_weights_other", group=new_group
    )
    yield (
        with_variable_weights_other,
        with_variable_weights_other_val,
        with_variable_weights_group,
        with_variable_weights.probs_and_group,
    )

    index_x = x.index(I[1])
    yield index_x, x_value[:, 1], x_probs_and_group, x.probs_and_group

    index_variable_weights = with_variable_weights.index(I[1])
    yield index_variable_weights, with_variable_weights_value[
        :, 1
    ], with_variable_weights_group, with_variable_weights.probs_and_group

    different_first_dim, different_first_dim_value, _ = make_discrete_var(
        (7, 42),
        17,
        "different_first_dim",
        group=x.probs_and_group,
    )
    yield different_first_dim, different_first_dim_value, x_probs_and_group, x.probs_and_group

    yield Concat(y, axis=0), torch.cat([y_value], dim=1), x_probs_and_group, x.probs_and_group

    yield Concat(y, y_dup, different_first_dim, axis=0), torch.cat(
        [y_value, y_dup_value, different_first_dim_value], dim=1
    ), x_probs_and_group, x.probs_and_group

    yield Concat.stack(y, axis=-1), torch.stack([y_value], dim=-1), x_probs_and_group, x.probs_and_group

    yield Concat.stack(y, y_dup, y, y, axis=-1), torch.stack(
        [y_value, y_dup_value, y_value, y_value], dim=-1
    ), x_probs_and_group, x.probs_and_group

    if just_mean:
        return

    full_mul_str = "a b, a c, d c a, d c a -> d a b c"
    full_mul = Einsum.from_einsum_string(full_mul_str, x, y_sums, z_rearrange, z_mul_const_rearrange)
    full_mul_eval = torch.einsum(
        "e a b, e a c, e d c a, e d c a -> e d a b c",
        x_value,
        y_sums_value,
        z_rearrange_eval,
        z_mul_const_rearrange_eval,
    )
    yield full_mul, full_mul_eval, x_probs_and_group, x.probs_and_group

    # for power in [1, 2, 3, 7, 8, 9, 13]:
    #     power_circuit = Einsum.power(z_rearrange, power=power, name=f"pow_{power}")
    #     yield power_circuit, z_rearrange_eval ** power, x_probs_and_group, x.probs_and_group

    # outer_product_x_y = Einsum.outer_product(x, y)
    # outer_product_x_y_eval = torch.einsum("e a b, e c d -> e a b c d", x_value, y_value)
    # yield outer_product_x_y, outer_product_x_y_eval, x_probs_and_group, x.probs_and_group

    # outer_product_x_y_const_r = Einsum.outer_product(x, y, const_r)
    # outer_product_x_y_const_r_eval = torch.einsum(
    #     "e a b, e c d, g h  -> e a b c d g h", x_value, y_value, const_r.value
    # )
    # yield outer_product_x_y_const_r, outer_product_x_y_const_r_eval, x_probs_and_group, x.probs_and_group

    # outer_product_x_y_x_x = Einsum.outer_product(x, y, x, x)
    # outer_product_x_y_x_x_eval = torch.einsum(
    #     "e a b, e c d, e g h, e o i  -> e a b c d g h o i", x_value, y_value, x_value, x_value
    # )
    # yield outer_product_x_y_x_x, outer_product_x_y_x_x_eval, x_probs_and_group, x.probs_and_group

    # outer_product_x_y_batch = Einsum.outer_product(x, y, name="x_y_outer_batch", num_batch_dims=1)
    # outer_product_x_y_batch_eval = torch.einsum("e a b, e a d -> e a b d", x_value, y_value)
    # yield outer_product_x_y_batch, outer_product_x_y_batch_eval, x_probs_and_group, x.probs_and_group

    # outer_product_x_y_batch_sum = Einsum.outer_product(
    #     x, y, name="x_y_outer_batch_sum", num_batch_dims=1, sum_over_batch=True
    # )
    # outer_product_x_y_batch_sum_eval = torch.einsum("e a b, e a d -> e b d", x_value, y_value)
    # yield outer_product_x_y_batch_sum, outer_product_x_y_batch_sum_eval, x_probs_and_group, x.probs_and_group

    # outer_product_x_y_x_x_batch = Einsum.outer_product(x, y, x, x, name="x_y_x_x_outer_batch", num_batch_dims=1)
    # outer_product_x_y_x_x_batch_eval = torch.einsum(
    #     "e a b, e a d, e a h, e a i -> e a b d h i", x_value, y_value, x_value, x_value
    # )
    # yield outer_product_x_y_x_x_batch, outer_product_x_y_x_x_batch_eval, x_probs_and_group, x.probs_and_group

    # outer_product_x_y_x_x_batch_sum = Einsum.outer_product(
    #     x, y, x, x, name="x_y_x_x_outer_batch_sum", num_batch_dims=1, sum_over_batch=True
    # )
    # outer_product_x_y_x_x_batch_sum_eval = torch.einsum(
    #     "e a b, e a d, e a h, e a i -> e b d h i", x_value, y_value, x_value, x_value
    # )
    # yield outer_product_x_y_x_x_batch_sum, outer_product_x_y_x_x_batch_sum_eval, x_probs_and_group, x.probs_and_group

    # variable_weights_mul_str = "a b, a -> b a"
    # variable_weights_mul = Einsum.from_einsum_str(
    #     variable_weights_mul_str, with_variable_weights, with_variable_weights_other
    # )
    # variable_weights_mul_eval = torch.einsum(
    #     "e a b, e a -> e b a", with_variable_weights_value, with_variable_weights_other_val
    # )
    # yield (
    #     variable_weights_mul,
    #     variable_weights_mul_eval,
    #     with_variable_weights_group,
    #     with_variable_weights.probs_and_group,
    # )

    # sigmoid_of = computational_node.sigmoid(full_mul)
    # sigmoid_of_eval = torch.sigmoid(full_mul_eval)
    # yield sigmoid_of, sigmoid_of_eval, x_probs_and_group, x.probs_and_group

    # sigmoid_of_variable = computational_node.sigmoid(variable_weights_mul)
    # sigmoid_of_variable_eval = torch.sigmoid(variable_weights_mul_eval)
    # yield (
    #     sigmoid_of_variable,
    #     sigmoid_of_variable_eval,
    #     with_variable_weights_group,
    #     with_variable_weights.probs_and_group,
    # )

    # # now for cumulants, oh baby

    # def empirical_mean_and_re_dim(x, group):
    #     return torch.broadcast_to(torch.einsum("i ..., i -> ...", x, group)[None], x.shape)

    # e_x = Cumulant((x,), name="e[x]")
    # e_x_val = empirical_mean_and_re_dim(x_value, x_probs_and_group)
    # yield e_x, e_x_val, x_probs_and_group, x.probs_and_group

    # e_variable_weights = Cumulant((with_variable_weights,), name="e[with_variable_weights]")
    # e_variable_weights_val = empirical_mean_and_re_dim(with_variable_weights_value, with_variable_weights_group)
    # yield e_variable_weights, e_variable_weights_val, with_variable_weights_group, with_variable_weights.probs_and_group

    # x_m_e_x = Add.from_unweighted_list([x, Einsum.scalar_mul(e_x, -1.0)])
    # x_m_e_x_val: torch.Tensor = x_value - e_x_val
    # yield x_m_e_x, x_m_e_x_val, x_probs_and_group, x.probs_and_group

    # # be wary, this code starts getting a bit cursed and insane right around here

    # variable_weights_m_e_variable_weights = Add.from_unweighted_list(
    #     [
    #         with_variable_weights,
    #         Einsum.scalar_mul(e_variable_weights, -1.0),
    #     ]
    # )
    # variable_weights_m_e_variable_weights_val: torch.Tensor = with_variable_weights_value - e_variable_weights_val
    # yield (
    #     variable_weights_m_e_variable_weights,
    #     variable_weights_m_e_variable_weights_val,
    #     with_variable_weights_group,
    #     with_variable_weights.probs_and_group,
    # )

    # centered_sqr_x = Einsum.power(x_m_e_x, 2)
    # centered_sqr_x_eval = x_m_e_x_val ** 2
    # yield centered_sqr_x, centered_sqr_x_eval, x_probs_and_group, x.probs_and_group

    # centered_sqr_variable_weights = Einsum.power(variable_weights_m_e_variable_weights, 2)
    # centered_sqr_variable_weights_eval = variable_weights_m_e_variable_weights_val ** 2
    # yield (
    #     centered_sqr_variable_weights,
    #     centered_sqr_variable_weights_eval,
    #     with_variable_weights_group,
    #     with_variable_weights.probs_and_group,
    # )

    # var_x = Cumulant((x, x), name="var[x]")
    # var_x_eval = empirical_mean_and_re_dim(
    #     torch.einsum("e a b, e c d -> e a b c d", x_m_e_x_val, x_m_e_x_val), x_probs_and_group
    # )
    # yield var_x, var_x_eval, x_probs_and_group, x.probs_and_group

    # var_variable_weights = Cumulant((with_variable_weights, with_variable_weights), name="var[with_variable_weights]")
    # var_variable_weights_eval = empirical_mean_and_re_dim(
    #     torch.einsum(
    #         "e a b, e c d -> e a b c d",
    #         variable_weights_m_e_variable_weights_val,
    #         variable_weights_m_e_variable_weights_val,
    #     ),
    #     with_variable_weights_group,
    # )
    # yield (
    #     var_variable_weights,
    #     var_variable_weights_eval,
    #     with_variable_weights_group,
    #     with_variable_weights.probs_and_group,
    # )

    # sigmoid_of_reduce = Einsum.from_einsum_str("d a b c -> b d", sigmoid_of)
    # sigmoid_of_reduce_eval = torch.einsum("e d a b c -> e b d", sigmoid_of_eval)
    # yield sigmoid_of_reduce, sigmoid_of_reduce_eval, x_probs_and_group, x.probs_and_group

    # new_dims, new_dims_value, _ = make_discrete_var((4, 5), "new_dims", 17, group=x.probs_and_group)
    # yield new_dims, new_dims_value, x_probs_and_group, x.probs_and_group

    # new_dims_1, new_dims_1_value, _ = make_discrete_var((6, 7), "new_dims_1", 17, group=x.probs_and_group)
    # yield new_dims_1, new_dims_1_value, x_probs_and_group, x.probs_and_group

    # c_new_dims: torch.Tensor = new_dims_value - empirical_mean_and_re_dim(new_dims_value, x_probs_and_group)
    # c_new_dims_1: torch.Tensor = new_dims_1_value - empirical_mean_and_re_dim(new_dims_1_value, x_probs_and_group)

    # third_cumulant_new_dims = Cumulant((x, new_dims, new_dims_1), name="new dims cum")

    # third_cumulant_new_dims_eval = empirical_mean_and_re_dim(
    #     torch.einsum("e a b, e c d, e g h -> e a b c d g h", x_m_e_x_val, c_new_dims, c_new_dims_1),
    #     x_probs_and_group,
    # )
    # yield third_cumulant_new_dims, third_cumulant_new_dims_eval, x_probs_and_group, x.probs_and_group

    # third_cumulant_simple_add = Cumulant((Add.from_unweighted_list([y, y_dup]),), name="cum simple add")

    # third_cumulant_simple_add_eval = empirical_mean_and_re_dim(y_sums_value, x_probs_and_group)
    # yield third_cumulant_simple_add, third_cumulant_simple_add_eval, x_probs_and_group, x.probs_and_group

    # for axis in range(0, 1):
    #     third_cumulant_simple_concat = Cumulant((Concat((y, y_dup), axis=axis),), name="cum simple cat")

    #     third_cumulant_simple_concat_eval = empirical_mean_and_re_dim(
    #         torch.cat([y_value, y_dup_value], dim=axis + 1),
    #         x_probs_and_group,
    #     )
    #     yield third_cumulant_simple_concat, third_cumulant_simple_concat_eval, x_probs_and_group, x.probs_and_group

    # for reverse in [False, True]:
    #     for is_concat, concat_axis in [(False, 0), (True, 0), (True, 1)]:
    #         a_items = (x,) * 2
    #         b_items = (new_dims,) * 3
    #         c_items = (new_dims_1,)
    #         if is_concat:
    #             a: Any = Concat(a_items, axis=concat_axis)
    #             b: Any = Concat(b_items, axis=concat_axis)
    #             c: Any = Concat(c_items, axis=concat_axis)
    #         else:
    #             a = Add.from_unweighted_list(a_items)
    #             b = Add.from_unweighted_list(b_items)
    #             c = Add.from_unweighted_list(c_items)
    #         third_cumulant_new_dims_op = Cumulant((c, b, a) if reverse else (a, b, c), name="new dims cum op")

    #         to_val_new_dims_op: Dict[Circuit, torch.Tensor] = {
    #             a: torch.cat([x_m_e_x_val] * 2, dim=concat_axis + 1) if is_concat else 2 * x_m_e_x_val,
    #             b: torch.cat([c_new_dims] * 3, dim=concat_axis + 1) if is_concat else 3 * c_new_dims,
    #             c: c_new_dims_1,
    #         }

    #         third_cumulant_new_dims_op_eval = empirical_mean_and_re_dim(
    #             torch.einsum(
    #                 "e a b, e c d, e g h -> e a b c d g h",
    #                 *(to_val_new_dims_op[c] for c in third_cumulant_new_dims_op.circuits),
    #             ),
    #             x_probs_and_group,
    #         )
    #         yield third_cumulant_new_dims_op, third_cumulant_new_dims_op_eval, x_probs_and_group, x.probs_and_group

    # a = Index(x, I[:, 1])
    # b = Index(new_dims, (3,))
    # c = Index(new_dims_1, I[:, -1])
    # third_cumulant_new_dims_index = Cumulant((a, b, c), name="new dims cum index")
    # third_cumulant_new_dims_index_eval = empirical_mean_and_re_dim(
    #     torch.einsum("e a, e c, e g -> e a c g", x_m_e_x_val[:, :, 1], c_new_dims[:, 3], c_new_dims_1[:, :, -1]),
    #     x_probs_and_group,
    # )
    # yield third_cumulant_new_dims_index, third_cumulant_new_dims_index_eval, x_probs_and_group, x.probs_and_group

    # cum_full_mul = Cumulant((full_mul,), "cum full mul")
    # yield (
    #     cum_full_mul,
    #     empirical_mean_and_re_dim(full_mul_eval, x_probs_and_group),
    #     x_probs_and_group,
    #     x.probs_and_group,
    # )

    # third_cumulant_x = Cumulant(
    #     (x, sigmoid_of_reduce, sigmoid_of_reduce),
    #     name="sigmoid cum",
    # )
    # c_sigmoid_x = sigmoid_of_reduce_eval - empirical_mean_and_re_dim(sigmoid_of_reduce_eval, x_probs_and_group)
    # third_cumulant_x_eval = empirical_mean_and_re_dim(
    #     torch.einsum("e a b, e c d, e g h -> e a b c d g h", x_m_e_x_val, c_sigmoid_x, c_sigmoid_x),
    #     x_probs_and_group,
    # )
    # yield third_cumulant_x, third_cumulant_x_eval, x_probs_and_group, x.probs_and_group

    # just_sig_3_cum_x = computational_node.sigmoid(third_cumulant_x)
    # just_sig_3_cum_x_eval = torch.sigmoid(third_cumulant_x_eval)
    # yield just_sig_3_cum_x, just_sig_3_cum_x_eval, x_probs_and_group, x.probs_and_group

    # sigmoid_third_cumulant_x = Einsum.from_einsum_str(
    #     f"a d, a b c d g h -> a b c d g h",
    #     x,
    #     just_sig_3_cum_x,
    #     name="sigmoid_3_cum_x",
    # )
    # sigmoid_third_cumulant_x_eval = torch.einsum(
    #     f"e a d, e a b c d g h -> e a b c d g h",
    #     x_value,
    #     just_sig_3_cum_x_eval,
    # )
    # yield sigmoid_third_cumulant_x, sigmoid_third_cumulant_x_eval, x_probs_and_group, x.probs_and_group

    # third_cumulant_variable = Cumulant(
    #     (with_variable_weights, sigmoid_of_variable, sigmoid_of_variable), name="sigmoid cum variable"
    # )
    # c_sigmoid_variable: torch.Tensor = sigmoid_of_variable_eval - empirical_mean_and_re_dim(
    #     sigmoid_of_variable_eval, with_variable_weights_group
    # )
    # third_cumulant_variable_eval = empirical_mean_and_re_dim(
    #     torch.einsum(
    #         "e a b, e c d, e g h -> e a b c d g h",
    #         variable_weights_m_e_variable_weights_val,
    #         c_sigmoid_variable,
    #         c_sigmoid_variable,
    #     ),
    #     with_variable_weights_group,
    # )
    # yield (
    #     third_cumulant_variable,
    #     third_cumulant_variable_eval,
    #     with_variable_weights_group,
    #     with_variable_weights.probs_and_group,
    # )

    # just_sig_3_cum_variable = computational_node.sigmoid(third_cumulant_variable)
    # just_sig_3_cum_variable_eval = torch.sigmoid(third_cumulant_variable_eval)
    # yield (
    #     just_sig_3_cum_variable,
    #     just_sig_3_cum_variable_eval,
    #     with_variable_weights_group,
    #     with_variable_weights.probs_and_group,
    # )

    # sigmoid_third_cumulant_variable = Einsum.from_einsum_str(
    #     f"a c, a b c d g h -> a b c d g h",
    #     with_variable_weights,
    #     just_sig_3_cum_variable,
    #     name="sigmoid_3_cum_variable",
    # )
    # sigmoid_third_cumulant_variable_eval = torch.einsum(
    #     f"e a c, e a b c d g h -> e a b c d g h",
    #     with_variable_weights_value,
    #     just_sig_3_cum_variable_eval,
    # )
    # yield (
    #     sigmoid_third_cumulant_variable,
    #     sigmoid_third_cumulant_variable_eval,
    #     with_variable_weights_group,
    #     with_variable_weights.probs_and_group,
    # )


def test_hash_seeder_randomness():
    assert default_hash_seeder()(Scalar(3.0)) != default_hash_seeder()(Scalar(3.0))
    assert default_hash_seeder(7)(Scalar(3.0)) == default_hash_seeder(7)(Scalar(3.0))
    assert default_hash_seeder(9)(Scalar(3.0)) == default_hash_seeder(9)(Scalar(3.0))

    torch.manual_seed(17)
    fst = default_hash_seeder()(Scalar(3.0))
    torch.manual_seed(17)
    snd = default_hash_seeder()(Scalar(3.0))
    assert fst == snd


def test_sample_randomness():
    x, x_value, x_probs_and_group = make_discrete_var((3, 2), 17, "x")

    def get_sampler(base_seed: Optional[int] = None):
        return Sampler(RandomSampleSpec((100,), seeder=default_hash_seeder(base_seed)))

    assert get_sampler().sample(x) != get_sampler().sample(x)
    assert torch.abs(get_sampler().sample(x).evaluate() - get_sampler().sample(x).evaluate()).mean() > 0.1

    assert get_sampler(2378).sample(x) == get_sampler(2378).sample(x)
    torch.testing.assert_allclose(get_sampler(2378).sample(x).evaluate(), get_sampler(2378).sample(x).evaluate())

    assert get_sampler().sample(x) != get_sampler().sample(x)
    assert torch.abs(get_sampler().sample(x).evaluate() - get_sampler().sample(x).evaluate()).mean() > 0.1

    def get_manual_sampler(seed: int = 1782, shape: tuple[int, ...] = (100,)):
        torch.manual_seed(seed)
        return Sampler(RandomSampleSpec(shape))

    assert get_manual_sampler().sample(x) == get_manual_sampler().sample(x)
    torch.testing.assert_allclose(get_manual_sampler().sample(x).evaluate(), get_manual_sampler().sample(x).evaluate())


# TODO: maybe this should be in a test_nb_rewrites.py or such?
def test_simplify_sampled():
    spec = RandomSampleSpec((100,), seeder=lambda _: 11)
    x = rc.DiscreteVar(rc.Array(torch.zeros((12, 37))))
    assert (
        (Sampler(spec).sample(x)).repr()
        == """0 [100,37] Index [tdf3c928b0b [100],:]
  1 [12,37] Array 4e751d26e9a4b2feca958f1a"""
    )
    spec.simplify = False
    assert (
        rc.strip_names_and_tags(Sampler(spec).sample(x)).repr()
        == """0 GeneralFunction gen_index_at_-2_batch_x_no_batch_index_c
  1 [12,37] Array 4e751d26e9a4b2feca958f1a
  2 GeneralFunction multinomial_[100]
    3 [12] Scalar 0.08333333333333333
    4 Tag 00000000-0000-0000-0000-000000000000
      'seed_11' [] Array 6287c955e43db41def49aa5f"""
    )


@pytest.mark.parametrize("replacement", [False, True])
def test_multinomial(replacement: bool):
    probs = torch.rand((5 * 6 * 7, 17 if replacement else 238))
    num_samples = 3 * 3 * 4
    seed = 17

    samps = torch.multinomial(
        probs, num_samples=num_samples, replacement=replacement, generator=torch.Generator().manual_seed(seed)
    )
    samp_circ = multinomial(Array(probs), Array(torch.tensor(seed)), shape=(num_samples,), replacement=replacement)
    samp_circ.print()
    assert Parser()(samp_circ.repr()) == samp_circ

    assert (samp_circ.evaluate() == samps).all()

    samp_circ = multinomial(
        Array(probs.reshape(5, 6 * 7, probs.shape[-1])),
        Array(torch.tensor(seed)),
        shape=(num_samples,),
        replacement=replacement,
    )
    samp_circ.print()
    assert Parser()(samp_circ.repr()) == samp_circ
    assert samp_circ.shape == (5, 6 * 7, num_samples)
    assert samp_circ.evaluate().shape == samp_circ.shape
    assert (samp_circ.evaluate().reshape(-1, num_samples) == samps).all()

    samp_circ = multinomial(
        Array(probs.reshape(5, 6 * 7, probs.shape[-1])),
        Array(torch.tensor(seed)),
        shape=(3, 4, 3),
        replacement=replacement,
    )
    samp_circ.print()
    assert Parser()(samp_circ.repr()) == samp_circ
    assert samp_circ.shape == (5, 6 * 7, 3, 4, 3)
    assert samp_circ.evaluate().shape == samp_circ.shape
    assert (samp_circ.evaluate().reshape(-1, num_samples) == samps).all()

    samp_circ = multinomial(
        Array(probs.reshape(5, 6, 7, probs.shape[-1])),
        Array(torch.tensor(seed)),
        shape=(3, 4, 3),
        replacement=replacement,
    )
    samp_circ.print()
    assert Parser()(samp_circ.repr()) == samp_circ
    assert samp_circ.shape == (5, 6, 7, 3, 4, 3)
    assert samp_circ.evaluate().shape == samp_circ.shape
    assert (samp_circ.evaluate().reshape(-1, num_samples) == samps).all()

    samp_circ = multinomial(
        Array(probs.reshape(5 * 6 * 7, probs.shape[-1])),
        Array(torch.tensor(seed)),
        shape=(3, 4, 3),
        replacement=replacement,
    )
    samp_circ.print()
    assert Parser()(samp_circ.repr()) == samp_circ
    assert samp_circ.shape == (5 * 6 * 7, 3, 4, 3)
    assert samp_circ.evaluate().shape == samp_circ.shape
    assert (samp_circ.evaluate().reshape(-1, num_samples) == samps).all()


@pytest.mark.parametrize("sample_shape", [(3,), (3, 4)])
def test_seed_sample(sample_shape):
    for v, _, *_ in make_discrete_var_test_circuits():
        # Check no error
        orig_shape = v.shape
        torch.manual_seed(1234)
        a1 = Sampler(RandomSampleSpec(sample_shape)).estimate_and_sample(v).evaluate()
        assert a1.shape == sample_shape + orig_shape
        torch.manual_seed(1234)
        a2 = Sampler(RandomSampleSpec(sample_shape)).estimate_and_sample(v).evaluate()
        torch.manual_seed(1234)
        a3 = Sampler(RandomSampleSpec(sample_shape, seeder=None)).estimate_and_sample(v).evaluate()
        torch.testing.assert_close(a1, a2)
        torch.testing.assert_close(a1, a3)
        torch.testing.assert_close(a2, a3)

        a1 = Sampler(RandomSampleSpec(sample_shape, seeder=default_hash_seeder(87))).estimate_and_sample(v).evaluate()
        assert a1.shape == sample_shape + orig_shape
        a2 = Sampler(RandomSampleSpec(sample_shape, seeder=default_hash_seeder(87))).estimate_and_sample(v).evaluate()
        torch.testing.assert_close(a1, a2)


@pytest.mark.parametrize("sample_shape", [(3,), (3, 4)])
def test_seed_per_circuit_sample(sample_shape):
    u, _, _ = make_discrete_var((3, 42), 17, "u")
    v, _, _ = make_discrete_var((3, 42), 17, "v")

    def seeder(c):
        if c == u.probs_and_group:
            return 0
        if c == v.probs_and_group:
            return 1
        return hash(c) ^ 2833

    c1 = Add(v, u)
    a1 = Sampler(RandomSampleSpec(sample_shape, seeder=seeder)).estimate_and_sample(c1).evaluate()
    c2 = Add(u, v)
    a2 = Sampler(RandomSampleSpec(sample_shape, seeder=seeder)).estimate_and_sample(c2).evaluate()
    torch.testing.assert_close(a1, a2)


def test_sample_all():
    for c, val, _, group_circuit in make_discrete_var_test_circuits():
        c.print()
        assert c.shape == val.shape[1:]
        sampled_circ = Sampler(RunDiscreteVarAllSpec([group_circuit])).estimate_and_sample(c)
        transform = sampled_circ.evaluate()
        torch.testing.assert_close(transform, val, atol=1e-5, rtol=1e-5)


def test_probs_dtype():
    int_vals = Array(torch.randint(0, 800, (100,)), name="int_vals")
    s = """
    'int_var' DiscreteVar
      'int_vals'
      'group' Tag 2ef9f1ac-28bb-4bed-ab7b-09b9050483ac
        'uniform_probs' [100] Scalar 0.01
    """

    c = Parser(reference_circuits_by_name=[int_vals]).parse_circuit(s)
    assert c.torch_dtype == torch.int64
    out = Sampler(RandomSampleSpec((8,))).estimate_and_sample(c)
    assert out.torch_dtype == torch.int64
    assert out.evaluate().dtype == torch.int64
    assert optimize_and_evaluate(out).dtype == torch.int64
